# pyOCD debugger
# Copyright (c) 2019-2020 Arm Limited
# SPDX-License-Identifier: Apache-2.0
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import pytest
import logging
from intervaltree import (Interval, IntervalTree)

from .conftest import mock

from pyocd.coresight.component import CoreSightCoreComponent
from pyocd.core import memory_map
from pyocd.coresight.rom_table import (
    CoreSightComponentID,
    ROMTable,
    )
from pyocd.core.memory_interface import MemoryInterface
from pyocd.debug.cache import MemoryCache
from pyocd.debug.context import DebugContext

class MockCoreForMemCache(CoreSightCoreComponent):
    """! @brief Just enough of a core to satisfy MemoryCache.
    
    Most importantly, it defines a memory map with a single RAM region covering almost the
    full 4 GB address space.
    """
    def __init__(self):
        self.run_token = 1
        self.memory_map = memory_map.MemoryMap(
            memory_map.RamRegion(name="ram", start=0, length=0xf0000000, cacheable=True)
            )

    def is_running(self):
        return False

MockDebugContext = mock.Mock(spec=DebugContext)

class RomMemory(MemoryCache, MemoryInterface):
    """! @brief Memory interface for reading constant values.
    
    Uses the memory cache as readily-available component to store data at fixed addresses. We
    just have to make sure the cache is never invalidated.
    """
    def __init__(self, ranges):
        """! @brief Constructor.
        
        @param self
        @param ranges Dict of start address -> list of word values.
        """
        super(RomMemory, self).__init__(MockDebugContext(), MockCoreForMemCache())
        
        # Fill in cache with data from ranges.
        for addr, data in ranges.items():
            self.write_memory_block32(addr, data)

    def _check_cache(self):
        # Ensure the cache is always valid.
        pass

class MockCoreSight(RomMemory):
    """! @brief RomMemory based on a list of MockCoreSightComponent objects."""
    def __init__(self, components):
        """! @brief Constructor.
        
        @param self
        @param components List of component dicts, where each component dict consists of start
            address -> list of word values.
        """
        ranges = {base: data for c in components for base, data in c.data.items()}
        super(MockCoreSight, self).__init__(ranges)
    
    @property
    def short_description(self):
        return "MockCoreSight"

class MockCoreSightComponent(object):
    """! @brief Generates a data dict from CoreSight component ID register values."""
    
    # Start offset within the 4 kB CoreSight component memory window of the ID registers
    # we care about, particularly those read by CoreSightComponentID.
    CMPID_REGS_OFFSET = 0xfbc
    
    def __init__(self, base, cidr, pidr, **kwargs):
        """! @brief Constructor.
        @param self
        @param base Base address of the component.
        @param cidr 32-bit combined CIDR register value.
        @param pidr 64-bit combined PIDR register value.
        @param kwargs Optional 'devarch', 'devtype', 'devid', and 'extra' keyword arguments. The
            first two are simple integers. 'devid' must be a list of 3 integers. 'extra' is a
            data dictionary (address -> list of 32-bit word values) for any extra fixed data that
            is contained in the component's 4 kB address space (such as a ROM table).
        """
        self._base = base
        self._pidr4_7 = (pidr >> 32)
        self._pidr0_3 = pidr & 0xffffffff
        self._cidr = cidr
        self._devarch = kwargs.get('devarch', 0)
        self._devid = kwargs.get('devid', [0, 0, 0])
        self._devtype = kwargs.get('devtype', 0)
        self._extra = kwargs.get('extra', {})
    
    @property
    def data(self):
        d = self._extra.copy()
        d.update({
            self._base + self.CMPID_REGS_OFFSET: [
                    self._devarch,                  # 0xfbc = DEVARCH
                    self._devid[2],                 # 0xfc0 = DEVID2
                    self._devid[1],                 # 0xfc4 = DEVID1
                    self._devid[0],                 # 0xfc8 = DEVID0
                    self._devtype,                  # 0xfcc = DEVTYPE
                    self._pidr4_7 & 0xff,           # 0xfd0 = PID4
                    (self._pidr4_7 >> 8) & 0xff,    # 0xfd4 = PID5
                    (self._pidr4_7 >> 16) & 0xff,   # 0xfd8 = PID6
                    (self._pidr4_7 >> 24) & 0xff,   # 0xfdc = PID7
                    self._pidr0_3 & 0xff,           # 0xfe0 = PID0
                    (self._pidr0_3 >> 8) & 0xff,    # 0xfe4 = PID1
                    (self._pidr0_3 >> 16) & 0xff,   # 0xfe8 = PID3
                    (self._pidr0_3 >> 24) & 0xff,   # 0xfec = PID3
                    self._cidr & 0xff,              # 0xff0 = CID0
                    (self._cidr >> 8) & 0xff,       # 0xff4 = CID1
                    (self._cidr >> 16) & 0xff,      # 0xff8 = CID2
                    (self._cidr >> 24) & 0xff,      # 0xffc = CID3
                ]
            })
        return d

class MockM4Components:
    """! @ brief Namespace for mock Cortex-M4 Class 0x1 ROM table and core complex components."""
    
    # ROM table #0 @ 0xe00ff000 (designer=244 part=00d)
    M4_ROM_TABLE_BASE = 0xe00ff000
    M4_ROM_TABLE = MockCoreSightComponent(M4_ROM_TABLE_BASE, cidr=0xb105100d, pidr=0x4000bb4c4,
        extra={
            # ROM table entries.
            0xe00ff000: [   0xfff0f003, # SCS
                            0xfff02003, # DWT
                            0xfff03003, # FPB
                            0xfff01003, # ITM
                            0xfff41003, # TPIU
                            0xfff42003, # ETM
                            0x00000000, 0x00000000, 0x00000000, 0x00000000, # (Terminator and extra)
                        ],
        })

    # [0]<e000e000:SCS-M4 class=14 designer=43b part=00c>
    SCS_BASE = 0xe000e000
    SCS = MockCoreSightComponent(SCS_BASE, cidr=0xb105e00d, pidr=0x4000bb00c)

    # [1]<e0001000:DWT class=14 designer=43b part=002>
    DWT_BASE = 0xe0001000
    DWT = MockCoreSightComponent(DWT_BASE, cidr=0xb105e00d, pidr=0x4003bb002)

    # [2]<e0002000:FPB class=14 designer=43b part=003>
    FPB_BASE = 0xe0002000
    FPB = MockCoreSightComponent(FPB_BASE, cidr=0xb105e00d, pidr=0x4002bb003)

    # [3]<e0000000:ITM class=14 designer=43b part=001>
    ITM_BASE = 0xe0000000
    ITM = MockCoreSightComponent(ITM_BASE, cidr=0xb105e00d, pidr=0x4003bb001)

    # [4]<e0040000:TPIU-M4 class=9 designer=43b part=9a1 devtype=11 archid=0000 devid=ca1:0:0>
    TPIU_BASE = 0xe0040000
    TPIU = MockCoreSightComponent(TPIU_BASE, cidr=0xb105900d, pidr=0x4000bb9a1, devtype=0x11, devid=[0xca1, 0, 0])

    # [5]<e0041000:ETM-M4 class=9 designer=43b part=925 devtype=13 archid=0000 devid=0:0:0>
    ETM_BASE = 0xe0041000
    ETM = MockCoreSightComponent(ETM_BASE, cidr=0xb105900d, pidr=0x4000bb925, devtype=0x13)
    
class MockCSSOC600Components:
    """! @ brief Namespace for mock Class 0x9 ROM table and CoreSight SoC-600 components."""
    
    C9_ROM_TABLE_BASE = 0x00000000
    C9_ROM_TABLE = MockCoreSightComponent(C9_ROM_TABLE_BASE, cidr=0xb105900d, pidr=0x4000bb7d5,
        devarch=0x47700af7, devid=[0x20, 0, 0],
        extra={
            # ROM table entries.
            0x00000000: [   0x00001003, # SDC-600
                            0x00002003, # SoC-600 AHB-AP
                            0x00000000, 0x00000000, 0x00000000, 0x00000000, # (Terminator and extra)
                            0x00000000, 0x00000000, 0x00000000, 0x00000000, # (extra)
                        ],
        })
    
    SDC600_BASE = 0x00001000
    SDC600 = MockCoreSightComponent(SDC600_BASE, cidr=0xb105900d, pidr=0x4000bb9ef, devarch=0x47700a57)
    
    C9_AHB_AP_BASE = 0x00002000
    C9_AHB_AP = MockCoreSightComponent(C9_AHB_AP_BASE, cidr=0xb105900d, pidr=0x4002bb9e3, devarch=0x47700a17)
    

# Complete set of components for a Cortex-M4 subsystem.
@pytest.fixture(scope='function')
def m4_rom():
    return MockCoreSight([
                MockM4Components.M4_ROM_TABLE,
                MockM4Components.SCS,
                MockM4Components.DWT,
                MockM4Components.FPB,
                MockM4Components.ITM,
                MockM4Components.TPIU,
                MockM4Components.ETM,
            ])

@pytest.fixture(scope='function')
def c9_top_rom():
    return MockCoreSight([
                MockCSSOC600Components.C9_ROM_TABLE,
                MockCSSOC600Components.SDC600,
                MockCSSOC600Components.C9_AHB_AP,
            ])

@pytest.fixture(scope='function')
def testrom():
    a = {
            0x1000: [1, 2, 3, 4],
            0x4000: [0xaa, 0xbeef],
            0x4400: [0xcccccccc],
        }
    return RomMemory(a)

@pytest.fixture(scope='function')
def testcoresight():
    return MockCoreSight([MockM4Components.M4_ROM_TABLE, MockM4Components.ETM])

class TestRomMemory:
    def test_r32(self, testrom):
        assert testrom.read32(0x1000) == 1
        assert testrom.read32(0x1004) == 2
        assert testrom.read32(0x4004) == 0xbeef
        assert testrom.read32(0x4400) == 0xcccccccc

    def test_r16(self, testrom):
        assert testrom.read16(0x1000) == 1
        assert testrom.read16(0x1002) == 0
        assert testrom.read16(0x4000) == 0xaa
        assert testrom.read16(0x4004) == 0xbeef
        assert testrom.read16(0x4402) == 0xcccc

    def test_r8(self, testrom):
        assert testrom.read8(0x1001) == 0
        assert testrom.read8(0x1003) == 0
        assert testrom.read8(0x1004) == 2
        assert testrom.read8(0x4001) == 0
        assert testrom.read8(0x4004) == 0xef
        assert testrom.read8(0x4005) == 0xbe
        assert testrom.read8(0x4401) == 0xcc

    def test_rb32(self, testrom):
        assert testrom.read_memory_block32(0x1000, 4) == [1, 2, 3, 4]
        assert testrom.read_memory_block32(0x4000, 2) == [0xaa, 0xbeef]

    def test_rb8(self, testrom):
        assert testrom.read_memory_block8(0x1008, 6) == [3, 0, 0, 0, 4, 0]
        assert testrom.read_memory_block8(0x4001, 6) == [0, 0, 0, 0xef, 0xbe, 00]
    
class TestMockCoreSight:
    def test_1(self, testcoresight):
        assert testcoresight.read32(0xe00ff000) == 0xfff0f003
        assert testcoresight.read32(0xe00ff000) == 0xfff0f003
        assert testcoresight.read_memory_block32(0xe00ffff0, 4) == [0xd, 0x10, 0x5, 0xb1]

class TestCoreSightComponentID:
    # Test parsing a non-CoreSight component in isolation.
    def test_scs(self):
        cmp = CoreSightComponentID(None, MockCoreSight([MockM4Components.SCS]),
                MockM4Components.SCS_BASE)
        cmp.read_id_registers()
        assert cmp.component_class == 14
        assert cmp.designer == 0x43b
        assert cmp.part == 0xc
        assert cmp.devarch == 0
        assert cmp.devid == [0, 0, 0]
    
    # Test parsing a CoreSight (class 9) component in isolation.
    def test_etm(self):
        cmp = CoreSightComponentID(None, MockCoreSight([MockM4Components.ETM]),
                MockM4Components.ETM_BASE)
        cmp.read_id_registers()
        assert cmp.component_class == 9
        assert cmp.designer == 0x43b
        assert cmp.part == 0x925
        assert cmp.devtype == 0x13
        assert cmp.archid == 0
        assert cmp.devarch == 0
        assert cmp.devid == [0, 0, 0]
    
    # Test parsing a CoreSight (class 9) component with a DEVID.
    def test_tpiu(self):
        cmp = CoreSightComponentID(None, MockCoreSight([MockM4Components.TPIU]),
                MockM4Components.TPIU_BASE)
        cmp.read_id_registers()
        assert cmp.component_class == 9
        assert cmp.designer == 0x43b
        assert cmp.part == 0x9a1
        assert cmp.devtype == 0x11
        assert cmp.archid == 0
        assert cmp.devarch == 0
        assert cmp.devid == [0xca1, 0, 0]
    
    # Test parsing a Class 0x9 ROM table.
    def test_c9_rom(self):
        cmp = CoreSightComponentID(None, MockCoreSight([MockCSSOC600Components.C9_ROM_TABLE]),
                MockCSSOC600Components.C9_ROM_TABLE_BASE)
        cmp.read_id_registers()
        assert cmp.component_class == 9
        assert cmp.is_rom_table

class TestRomTable:
    # Test a Class 0x1 ROM table.
    def test_m4_rom(self, m4_rom):
        # Read ROM table component ID.
        cmpid = CoreSightComponentID(None, m4_rom, MockM4Components.M4_ROM_TABLE_BASE)
        cmpid.read_id_registers()
        
        # Create the ROM table.
        rom_table = ROMTable.create(m4_rom, cmpid)
        rom_table.init()
        
        # Verify all components were parsed.
        assert len(rom_table.components) == 6
        
        # Check SCS-M4.
        scs = rom_table.components[0]
        assert scs.component_class == 14
        assert scs.designer == 0x43b
        assert scs.part == 0xc
        
        # Check TPIU.
        tpiu = rom_table.components[4]
        assert tpiu.component_class == 9
        assert tpiu.part == 0x9a1
        assert tpiu.devid == [0xca1, 0, 0]
        
    # Test a Class 0x9 ROM table and CS-600 components.
    def test_c9_rom(self, c9_top_rom):
        # Read ROM table component ID.
        cmpid = CoreSightComponentID(None, c9_top_rom, MockCSSOC600Components.C9_ROM_TABLE_BASE)
        cmpid.read_id_registers()
        
        # Create the ROM table.
        rom_table = ROMTable.create(c9_top_rom, cmpid)
        rom_table.init()

        # Validate ROM table properties.
        assert rom_table._width == 32
        assert not rom_table.has_com_port
        assert rom_table.has_prr
        assert not rom_table.is_sysmem

        # Validate components.
        assert len(rom_table.components) == 2
        
        # Validate SDC-600.
        sdc = rom_table.components[0]
        assert sdc.component_class == 9
        assert sdc.designer == 0x43b
        assert sdc.part == 0x9ef
        assert sdc.archid == 0xa57
        
        # Validate AHB-AP.
        ahb = rom_table.components[1]
        assert ahb.component_class == 9
        assert ahb.designer == 0x43b
        assert ahb.part == 0x9e3
        assert ahb.archid == 0xa17




# pyOCD debugger
# Copyright (c) 2015-2020 Arm Limited
# Copyright (c) 2021 Chris Reed
# SPDX-License-Identifier: Apache-2.0
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import logging
from enum import Enum
from typing import NamedTuple

from ..core import (exceptions, memory_interface)
from ..core.target import Target
from ..probe.debug_probe import DebugProbe
from ..probe.swj import SWJSequenceSender
from .ap import APSEL_APBANKSEL
from ..utility.sequencer import CallSequence
from ..utility.timeout import Timeout

LOG = logging.getLogger(__name__)

TRACE = LOG.getChild("trace")
TRACE.setLevel(logging.CRITICAL)

# DP register addresses. The DPBANKSEL value is encoded in bits [7:4].
DP_IDR = 0x00 # read-only
DP_IDR1 = 0x10 # read-only
DP_BASEPTR0 = 0x20 # read-only
DP_BASEPTR1 = 0x30 # read-only
DP_ABORT = 0x00 # write-only
DP_CTRL_STAT = 0x04 # read-write
DP_DLCR = 0x14 # read-write
DP_TARGETID = 0x24 # read-only
DP_DLPIDR = 0x34 # read-only
DP_EVENTSTAT = 0x44 # read-only
DP_SELECT1 = 0x54 # write-only
DP_SELECT = 0x8 # write-only
DP_RDBUFF = 0xC # read-only

# Mask and shift for extracting DPBANKSEL from our DP register address constants. These are not
# related to the SELECT.DPBANKSEL bitfield.
DPADDR_MASK = 0x0f
DPADDR_DPBANKSEL_MASK = 0xf0
DPADDR_DPBANKSEL_SHIFT = 4

DPIDR1_ASIZE_MASK = 0x00000007f
DPIDR1_ERRMODE_MASK = 0x00000080

BASEPTR0_VALID_MASK = 0x00000001
BASEPTR0_PTR_MASK = 0xfffff000
BASEPTR0_PTR_SHIFT = 12

ABORT_DAPABORT = 0x00000001
ABORT_STKCMPCLR = 0x00000002
ABORT_STKERRCLR = 0x00000004
ABORT_WDERRCLR = 0x00000008
ABORT_ORUNERRCLR = 0x00000010

# DP Control / Status Register bit definitions
CTRLSTAT_ORUNDETECT = 0x00000001
CTRLSTAT_STICKYORUN = 0x00000002
CTRLSTAT_STICKYCMP = 0x00000010
CTRLSTAT_STICKYERR = 0x00000020
CTRLSTAT_READOK = 0x00000040
CTRLSTAT_WDATAERR = 0x00000080

# DP SELECT register fields.
SELECT_DPBANKSEL_MASK = 0x0000000f
SELECT_APADDR_MASK = 0xfffffff0

DPIDR_REVISION_MASK = 0xf0000000
DPIDR_REVISION_SHIFT = 28
DPIDR_PARTNO_MASK = 0x0ff00000
DPIDR_PARTNO_SHIFT = 20
DPIDR_MIN_MASK = 0x00010000
DPIDR_VERSION_MASK = 0x0000f000
DPIDR_VERSION_SHIFT = 12

CSYSPWRUPACK = 0x80000000
CDBGPWRUPACK = 0x20000000
CSYSPWRUPREQ = 0x40000000
CDBGPWRUPREQ = 0x10000000

TRNNORMAL = 0x00000000
MASKLANE = 0x00000f00

## Arbitrary 5 second timeout for DP power up/down requests.
DP_POWER_REQUEST_TIMEOUT = 5.0

## @brief Class to hold fields from DP IDR register.
class DPIDR(NamedTuple):
    idr: int
    partno: int
    version: int
    revision: int
    mindp: int

class ADIVersion(Enum):
    """! @brief Supported versions of the Arm Debug Interface."""
    ADIv5 = 5
    ADIv6 = 6

class DPConnector:
    """! @brief Establishes a connection to the DP for a given wire protocol.
    
    This class will ask the probe to connect using a given wire protocol. Then it makes multiple
    attempts at sending the SWJ sequence to select the wire protocol and read the DP IDR register.
    """
    
    def __init__(self, probe):
        self._probe = probe
        self._session = probe.session
        self._idr = None
        
        # Make sure we have a session, since we get the session from the probe and probes have their session set
        # after creation.
        assert self._session is not None, "DPConnector requires the probe to have a session"
    
    @property
    def idr(self):
        """! @brief DPIDR instance containing values read from the DP IDR register."""
        return self._idr
    
    def _get_protocol(self, protocol):
        # Convert protocol from setting if not passed as parameter.
        if protocol is None:
            protocol_name = self._session.options.get('dap_protocol').strip().lower()
            protocol = DebugProbe.PROTOCOL_NAME_MAP[protocol_name]
            if protocol not in self._probe.supported_wire_protocols:
                raise exceptions.DebugError("requested wire protocol %s not supported by the debug probe" % protocol.name)
        return protocol
    
    def connect(self, protocol=None):
        """! @brief Establish a connection to the DP.
        
        This method causes the debug probe to connect using the wire protocol.
        
        @param self
        @param protocol One of the @ref pyocd.probe.debug_probe.DebugProbe.Protocol
            "DebugProbe.Protocol" enums. If not provided, will default to the `dap_protocol` setting.
        
        @exception DebugError
        @exception TransferError
        """
        try:
            self._probe.lock()

            # Determine the requested wire protocol.
            protocol = self._get_protocol(protocol)

            # If this is not None then the probe is already connected.
            current_wire_protocol = self._probe.wire_protocol
            already_connected = current_wire_protocol is not None
        
            if already_connected:
                self._check_protocol(current_wire_protocol, protocol)
            else:
                self._connect_probe(protocol)

            protocol = self._probe.wire_protocol
            self._connect_dp(protocol)
        finally:
            self._probe.unlock()
    
    def _check_protocol(self, current_wire_protocol, protocol):
        # Warn about mismatched current and requested wire protocols.
        if (protocol is not current_wire_protocol) and (protocol is not DebugProbe.Protocol.DEFAULT):
            LOG.warning("Cannot use %s; already connected with %s", protocol.name, current_wire_protocol.name)
        else:
            LOG.debug("Already connected with %s", current_wire_protocol.name)

    def _connect_probe(self, protocol):
        # Debug log with the selected protocol.
        if protocol is not DebugProbe.Protocol.DEFAULT:
            LOG.debug("Using %s wire protocol", protocol.name)
        
        # Connect using the selected protocol.
        self._probe.connect(protocol)

        # Log the actual protocol if selected was default.
        if protocol is DebugProbe.Protocol.DEFAULT:
            protocol = self._probe.wire_protocol
            LOG.debug("Default wire protocol selected; using %s", protocol.name)
        
    def _connect_dp(self, protocol):
        # Get SWJ settings.
        use_dormant = self._session.options.get('dap_swj_use_dormant')
        send_swj = self._session.options.get('dap_swj_enable') \
                and (DebugProbe.Capability.SWJ_SEQUENCE in self._probe.capabilities)

        # Create object to send SWJ sequences.
        swj = SWJSequenceSender(self._probe, use_dormant)
        
        # Multiple attempts to select protocol and read DP IDR.
        for attempt in range(4):
            try:
                if send_swj:
                    swj.select_protocol(protocol)
                
                # Attempt to read the DP IDR register.
                self._idr = self.read_idr()
                
                # Successful connection so exit the loop.
                break
            except exceptions.TransferError:
                # If not sending the SWJ sequence, just reraise; there's nothing more to do.
                if not send_swj:
                    raise
                
                # If the read of the DP IDCODE fails, retry SWJ sequence. The DP may have been
                # in a state where it thought the SWJ sequence was an invalid transfer. We also
                # try enabling use of dormant state if it wasn't already enabled.
                LOG.debug("DP IDCODE read failed; resending SWJ sequence (use dormant=%s)", use_dormant)
                
                if attempt == 1:
                    # If already using dormant mode, just raise, we don't need to retry the same mode.
                    if use_dormant:
                        raise
                    
                    # After the second attempt, switch to enabling dormant mode.
                    swj.use_dormant = True
                elif attempt == 3:
                    # After 4 attempts, we let the exception propagate.
                    raise

    def read_idr(self):
        """! @brief Read IDR register and get DP version"""
        dpidr = self._probe.read_dp(DP_IDR, now=True)
        dp_partno = (dpidr & DPIDR_PARTNO_MASK) >> DPIDR_PARTNO_SHIFT
        dp_version = (dpidr & DPIDR_VERSION_MASK) >> DPIDR_VERSION_SHIFT
        dp_revision = (dpidr & DPIDR_REVISION_MASK) >> DPIDR_REVISION_SHIFT
        is_mindp = (dpidr & DPIDR_MIN_MASK) != 0
        return DPIDR(dpidr, dp_partno, dp_version, dp_revision, is_mindp)

class DebugPort:
    """! @brief Represents the Arm Debug Interface (ADI) Debug Port (DP)."""

    ## Sleep for 50 ms between connection tests and reconnect attempts after a reset.
    _RESET_RECOVERY_SLEEP_INTERVAL = 0.05

    ## Number of times to try to read DP registers after hw reset before attempting reconnect.
    _RESET_RECOVERY_ATTEMPTS_BEFORE_RECONNECT = 1
    
    def __init__(self, probe, target):
        """! @brief Constructor.
        @param self The DebugPort object.
        @param probe The @ref pyocd.probe.debug_probe.DebugProbe "DebugProbe" object. The probe is assumed to not
            have been opened yet.
        @param target An instance of @ref pyocd.core.soc_target.SoCTarget "SoCTarget". Assumed to not have been
            fully initialized.
        """
        self._probe = probe
        self.target = target
        self._session = target.session
        self.valid_aps = None
        self.dpidr = None
        self.aps = {}
        self._access_number = 0
        self._cached_dp_select = None
        self._protocol = None
        self._probe_managed_ap_select = False
        self._probe_managed_dpbanksel = False
        self._probe_supports_dpbanksel = False
        self._probe_supports_apv2_addresses = False
        self._have_probe_capabilities = False
        self._did_check_version = False
        self._log_dp_info = True
        
        # DPv3 attributes
        self._is_dpv3 = False
        self._addr_size = None
        self._addr_mask = None
        self._errmode = None
        self._base_addr = None
        self._apacc_mem_interface = None
        
        # Subscribe to reset events.
        self._session.subscribe(self._reset_did_occur, (Target.Event.PRE_RESET, Target.Event.POST_RESET))

    @property
    def probe(self):
        return self._probe
    
    @property
    def session(self):
        return self._session
    
    @property
    def adi_version(self):
        return ADIVersion.ADIv6 if self._is_dpv3 else ADIVersion.ADIv5
    
    @property
    def base_address(self):
        """! @brief Base address of the first component for an ADIv6 system."""
        return self._base_addr
    
    @property
    def apacc_memory_interface(self):
        """! @brief Memory interface for performing APACC transactions."""
        if self._apacc_mem_interface is None:
            self._apacc_mem_interface = APAccessMemoryInterface(self)
        return self._apacc_mem_interface
    
    @property
    def next_access_number(self):
        self._access_number += 1
        return self._access_number
    
    def lock(self):
        """! @brief Lock the DP from access by other threads."""
        self.probe.lock()
    
    def unlock(self):
        """! @brief Unlock the DP."""
        self.probe.unlock()

    def connect(self, protocol=None):
        """! @brief Connect to the target.
        
        This method causes the debug probe to connect using the selected wire protocol. The probe
        must have already been opened prior to this call.
        
        Unlike create_connect_sequence(), this method is intended to be used when manually constructing a
        DebugPort instance. It simply calls create_connect_sequence() and invokes the returned call sequence.
        
        @param self
        @param protocol One of the @ref pyocd.probe.debug_probe.DebugProbe.Protocol
            "DebugProbe.Protocol" enums. If not provided, will default to the `protocol` setting.
        """
        self._protocol = protocol
        self.create_connect_sequence().invoke()
    
    def disconnect(self):
        """! @brief Disconnect from target.
        
        DP debug is powered down. See power_down_debug().
        """
        self.power_down_debug()

    def create_connect_sequence(self):
        """! @brief Returns call sequence to connect to the target.
        
        Returns a @ref pyocd.utility.sequence.CallSequence CallSequence that will connect to the
        DP, power up debug and the system, check the DP version to identify whether the target uses
        ADI v5 or v6, then clears sticky errors.
        
        The probe must have already been opened prior to this method being called.
        
        @param self
        @return @ref pyocd.utility.sequence.CallSequence CallSequence
        """
        seq = [
            ('lock_probe',          self.probe.lock),
            ]
        if not self._have_probe_capabilities:
            seq += [
                ('get_probe_capabilities', self._get_probe_capabilities),
                ]
        seq += [
            ('connect',             self._connect),
            ('clear_sticky_err',    self.clear_sticky_err),
            ('power_up_debug',      self.power_up_debug),
            ]
        if not self._did_check_version:
            seq += [
                ('check_version',       self._check_version),
                ]
        seq += [
            ('unlock_probe',        self.probe.unlock),
            ]
        return CallSequence(*seq)

    def _get_probe_capabilities(self):
        """! @brief Examine the probe's capabilities."""
        caps = self._probe.capabilities
        self._probe_managed_ap_select = (DebugProbe.Capability.MANAGED_AP_SELECTION in caps)
        self._probe_managed_dpbanksel = (DebugProbe.Capability.MANAGED_DPBANKSEL in caps)
        self._probe_supports_dpbanksel = (DebugProbe.Capability.BANKED_DP_REGISTERS in caps)
        self._probe_supports_apv2_addresses = (DebugProbe.Capability.APv2_ADDRESSES in caps)
        self._have_probe_capabilities = True

    def _connect(self):
        # Attempt to connect.
        connector = DPConnector(self.probe)
        connector.connect(self._protocol)

        # Report on DP version.
        self.dpidr = connector.idr
        LOG.log(logging.INFO if self._log_dp_info else logging.DEBUG,
            "DP IDR = 0x%08x (v%d%s rev%d)", self.dpidr.idr, self.dpidr.version,
            " MINDP" if self.dpidr.mindp else "", self.dpidr.revision)
        
    def _check_version(self):
        self._is_dpv3 = (self.dpidr.version == 3)
        if self._is_dpv3:
            # Check that the probe will be able to access ADIv6 APs.
            if self._probe_managed_ap_select and not self._probe_supports_apv2_addresses:
                raise exceptions.ProbeError("connected to ADIv6 target with probe that does not support APv2 addresses")
            
            idr1 = self.read_reg(DP_IDR1)
            
            self._addr_size = idr1 & DPIDR1_ASIZE_MASK
            self._addr_mask = (1 << self._addr_size) - 1
            self._errmode_supported = (idr1 & DPIDR1_ERRMODE_MASK) != 0
            
            LOG.debug("DP IDR1 = 0x%08x (addr size=%d, errmode=%d)", idr1, self._addr_size, self._errmode_supported)
            
            # Read base system address.
            baseptr0 = self.read_reg(DP_BASEPTR0)
            valid = (baseptr0 & BASEPTR0_VALID_MASK) != 0
            if valid:
                base = (baseptr0 & BASEPTR0_PTR_MASK) >> BASEPTR0_PTR_SHIFT
                if self._addr_size > 32:
                    baseptr1 = self.read_reg(DP_BASEPTR1)
                    base |= baseptr1 << 32

                base &= self._addr_mask
                self._base_addr = base
                
                LOG.debug("DP BASEPTR = 0x%08x", self._base_addr)
            else:
                LOG.warning("DPv3 has no valid base address")

        self._did_check_version = True

    def flush(self):
        try:
            self.probe.flush()
        except exceptions.TargetError as error:
            self._handle_error(error, self.next_access_number)
            raise

    def read_reg(self, addr, now=True):
        return self.read_dp(addr, now)

    def write_reg(self, addr, data):
        self.write_dp(addr, data)

    def power_up_debug(self):
        """! @brief Assert DP power requests.
        
        Request both debug and system power be enabled, and wait until the request is acked.
        There is a timeout for the request.
        
        @return Boolean indicating whether the power up request succeeded.
        """
        # Send power up request for system and debug.
        self.write_reg(DP_CTRL_STAT, CSYSPWRUPREQ | CDBGPWRUPREQ | MASKLANE | TRNNORMAL)

        with Timeout(DP_POWER_REQUEST_TIMEOUT) as time_out:
            while time_out.check():
                r = self.read_reg(DP_CTRL_STAT)
                if (r & (CDBGPWRUPACK | CSYSPWRUPACK)) == (CDBGPWRUPACK | CSYSPWRUPACK):
                    break
            else:
                return False
        
        return True

    def power_down_debug(self):
        """! @brief Deassert DP power requests.
        
        ADIv6 says that we must not clear CSYSPWRUPREQ and CDBGPWRUPREQ at the same time.
        ADIv5 says CSYSPWRUPREQ must not be set to 1 while CDBGPWRUPREQ is set to 0. So we
        start with deasserting system power, then debug power. Each deassertion has its own
        timeout.
        
        @return Boolean indicating whether the power down request succeeded.
        """
        # Power down system first.
        self.write_reg(DP_CTRL_STAT, CDBGPWRUPREQ | MASKLANE | TRNNORMAL)
        
        with Timeout(DP_POWER_REQUEST_TIMEOUT) as time_out:
            while time_out.check():
                r = self.read_reg(DP_CTRL_STAT)
                if (r & (CDBGPWRUPACK | CSYSPWRUPACK)) == CDBGPWRUPACK:
                    break
            else:
                return False

        # Now power down debug.
        self.write_reg(DP_CTRL_STAT,  MASKLANE | TRNNORMAL)
        
        with Timeout(DP_POWER_REQUEST_TIMEOUT) as time_out:
            while time_out.check():
                r = self.read_reg(DP_CTRL_STAT)
                if (r & (CDBGPWRUPACK | CSYSPWRUPACK)) == 0:
                    break
            else:
                return False
        
        return True

    def _invalidate_cache(self):
        """! @brief Invalidate cached DP registers."""
        self._cached_dp_select = None
    
    def _reset_did_occur(self, notification):
        """! @brief Handles reset notifications to invalidate register cache.
        
        The cache is cleared on all resets just to be safe. On most devices, warm resets do not reset
        debug logic, but it does happen on some devices.
        """
        self._invalidate_cache()
    
    def post_reset_recovery(self):
        """! @brief Wait for the target to recover from reset, with auto-reconnect if needed."""
        # Check if we can access DP registers. If this times out, then reconnect the DP and retry.
        with Timeout(self.session.options.get('reset.dap_recover.timeout'),
                self._RESET_RECOVERY_SLEEP_INTERVAL) as time_out:
            attempt = 0
            while time_out.check():
                try:
                    # Try to read CTRL/STAT. If the power-up bits request are reset, then the DP
                    # connection was not lost and we can just return.
                    value = self.read_reg(DP_CTRL_STAT)
                    if (value & (CSYSPWRUPREQ | CDBGPWRUPREQ)) == (CSYSPWRUPREQ | CDBGPWRUPREQ):
                        return
                except exceptions.TransferError:
                    # Ignore errors caused by flushing.
                    try:
                        self.flush()
                    except exceptions.TransferError:
                        pass

                if attempt == self._RESET_RECOVERY_ATTEMPTS_BEFORE_RECONNECT:
                    LOG.info("DAP is not accessible after reset; attempting reconnect")
                elif attempt > self._RESET_RECOVERY_ATTEMPTS_BEFORE_RECONNECT:
                    # Try reconnect.
                    try:
                        self._log_dp_info = False
                        self.connect()
                    finally:
                        self._log_dp_info = True

                attempt += 1
            else:
                LOG.error("DAP is not accessible after reset followed by attempted reconnect")

    def reset(self, *, send_notifications=True):
        """! @brief Hardware reset.
        
        Pre- and post-reset notifications are sent.
        
        This method can be called before the DebugPort is connected.

        @param self This object.
        @param send_notifications Optional keyword-only parameter used by higher-level reset methods so they can
            manage the sending of reset notifications themselves, in order to provide more context in the notification.
        
        @todo Should automatic recovery from a disconnected DAP be provided for these low-level hardware resets
            like is done for CortexM.reset()?
        """
        if send_notifications:
            self.session.notify(Target.Event.PRE_RESET, self)

        self.probe.reset()
        self.post_reset_recovery()

        if send_notifications:
            self.session.notify(Target.Event.POST_RESET, self)

    def assert_reset(self, asserted, *, send_notifications=True):
        """! @brief Assert or deassert the hardware reset signal.
        
        A pre-reset notification is sent before asserting reset, whereas a post-reset notification is sent
        after deasserting reset.
        
        This method can be called before the DebugPort is connected.
        
        @param self This object.
        @param asserted True if nRESET is to be driven low; False will drive nRESET high.
        @param send_notifications Optional keyword-only parameter used by higher-level reset methods so they can
            manage the sending of reset notifications themselves, in order to provide more context in the notification.
        """
        is_asserted = False
        if send_notifications:
            is_asserted = self.is_reset_asserted()
            if asserted and not is_asserted:
                self.session.notify(Target.Event.PRE_RESET, self)

        self.probe.assert_reset(asserted)

        if send_notifications and not asserted and is_asserted:
            self.session.notify(Target.Event.POST_RESET, self)

    def is_reset_asserted(self):
        """! @brief Returns the current state of the nRESET signal.
        
        This method can be called before the DebugPort is initalized.

        @retval True Reset is asserted; nRESET is low.
        @retval False Reset is not asserted; nRESET is high.
        """
        return self.probe.is_reset_asserted()

    def set_clock(self, frequency):
        """! @brief Change the wire protocol's clock frequency.
        @param self This object.
        @param frequency New wire protocol frequency in Hertz.
        """
        self.probe.set_clock(frequency)

    def _write_dp_select(self, mask, value):
        """! @brief Modify part of the DP SELECT register and write if cache is stale.
        
        The DP lock must already be acquired before calling this method.
        """
        # Compute the new SELECT value and see if we need to write it.
        if self._cached_dp_select is None:
            select = value
        else:
            select = (self._cached_dp_select & ~mask) | value
            if select == self._cached_dp_select:
                return
        
        # Update the SELECT register and cache.
        self.write_dp(DP_SELECT, select)
        self._cached_dp_select = select
    
    def _set_dpbanksel(self, addr, is_write):
        """! @brief Updates the DPBANKSEL field of the SELECT register as required.
        
        Several DP registers (most, actually) ignore DPBANKSEL. If one of those is being
        accessed, any value of DPBANKSEL can be used. Otherwise SELECT is updated if necessary
        and a lock acquired so another thread doesn't change DPBANKSEL until thsi transaction is
        complete.
        
        This method also handles the case where the debug probe manages DPBANKSEL on its own,
        such as with STLink.
        
        @return Whether the access needs a lock on DP SELECT.
        @exception exceptions.ProbeError Raised when a banked register is being accessed but the
            probe doesn't support DPBANKSEL.
        """
        # For DPv1-2, only address 0x4 (CTRL/STAT) honours DPBANKSEL.
        # For DPv3, SELECT and RDBUFF ignore DPBANKSEL for both reads and writes, while
        # ABORT ignores it only for writes (address 0 for reads is IDR).
        if self._is_dpv3 and not is_write:
            registers_ignoring_dpbanksel = (DP_SELECT, DP_RDBUFF)
        else:
            registers_ignoring_dpbanksel = (DP_ABORT, DP_SELECT, DP_RDBUFF)
        
        if (addr & DPADDR_MASK) not in registers_ignoring_dpbanksel:
            # Get the DP bank.
            dpbanksel = (addr & DPADDR_DPBANKSEL_MASK) >> DPADDR_DPBANKSEL_SHIFT
            
            # Check if the probe handles this for us.
            if self._probe_managed_dpbanksel:
                # If there is a nonzero DPBANKSEL and the probe doesn't support this,
                # then report an error.
                if dpbanksel and not self._probe_supports_dpbanksel:
                    raise exceptions.ProbeError("probe does not support banked DP registers")
                else:
                    return False
            
            # Update the selected DP bank.
            self.lock()
            self._write_dp_select(SELECT_DPBANKSEL_MASK, dpbanksel)
            return True
        else:
            return False

    def read_dp(self, addr, now=True):
        if (addr & DPADDR_MASK) % 4 != 0:
            raise ValueError("DP address must be word aligned")
        num = self.next_access_number
        
        # Update DPBANKSEL if required.
        did_lock = self._set_dpbanksel(addr, False)

        try:
            result_cb = self.probe.read_dp(addr & DPADDR_MASK, now=False)
        except exceptions.TargetError as error:
            self._handle_error(error, num)
            if did_lock:
                self.unlock()
            raise
        except Exception:
            if did_lock:
                self.unlock()
            raise

        # Read callback returned for async reads.
        def read_dp_cb():
            try:
                result = result_cb()
                TRACE.debug("read_dp:%06d %s(addr=0x%08x) -> 0x%08x", num, "" if now else "...", addr, result)
                return result
            except exceptions.TargetError as error:
                TRACE.debug("read_dp:%06d %s(addr=0x%08x) -> error (%s)", num, "" if now else "...", addr, error)
                self._handle_error(error, num)
                raise
            finally:
                if did_lock:
                    self.unlock()

        if now:
            return read_dp_cb()
        else:
            TRACE.debug("read_dp:%06d (addr=0x%08x) -> ...", num, addr)
            return read_dp_cb

    def write_dp(self, addr, data):
        if (addr & DPADDR_MASK) % 4 != 0:
            raise ValueError("DP address must be word aligned")
        num = self.next_access_number
        
        # Update DPBANKSEL if required.
        did_lock = self._set_dpbanksel(addr, True)

        # Write the DP register.
        try:
            TRACE.debug("write_dp:%06d (addr=0x%08x) = 0x%08x", num, addr, data)
            self.probe.write_dp(addr & DPADDR_MASK, data)
        except exceptions.TargetError as error:
            self._handle_error(error, num)
            raise
        finally:
            if did_lock:
                self.unlock()

        return True
    
    def _select_ap(self, addr):
        """! @brief Write DP_SELECT to choose the given AP.
        
        Handles the case where the debug probe manages selecting an AP itself, in which case we
        never write SELECT directly.

        @return Whether the access needs a lock on DP SELECT.
        """
        # If the probe handles selecting the AP for us, there's nothing to do here.
        if self._probe_managed_ap_select:
            return False
        
        # Write DP SELECT to select the probe.
        self.lock()
        if self.adi_version == ADIVersion.ADIv5:
            self._write_dp_select(APSEL_APBANKSEL, addr & APSEL_APBANKSEL)
        elif self.adi_version == ADIVersion.ADIv6:
            self._write_dp_select(SELECT_APADDR_MASK, addr & SELECT_APADDR_MASK)
        else:
            self.unlock()
            assert False, "invalid ADI version"
        return True

    def write_ap(self, addr, data):
        assert isinstance(addr, int)
        num = self.next_access_number
        did_lock = False

        try:
            did_lock = self._select_ap(addr)
            TRACE.debug("write_ap:%06d (addr=0x%08x) = 0x%08x", num, addr, data)
            self.probe.write_ap(addr, data)
        except exceptions.TargetError as error:
            self._handle_error(error, num)
            raise
        finally:
            if did_lock:
                self.unlock()

        return True

    def read_ap(self, addr, now=True):
        assert isinstance(addr, int)
        num = self.next_access_number
        did_lock = False

        try:
            did_lock = self._select_ap(addr)
            result_cb = self.probe.read_ap(addr, now=False)
        except exceptions.TargetError as error:
            self._handle_error(error, num)
            if did_lock:
                self.unlock()
            raise
        except Exception:
            if did_lock:
                self.unlock()
            raise

        # Read callback returned for async reads.
        def read_ap_cb():
            try:
                result = result_cb()
                TRACE.debug("read_ap:%06d %s(addr=0x%08x) -> 0x%08x", num, "" if now else "...", addr, result)
                return result
            except exceptions.TargetError as error:
                TRACE.debug("read_ap:%06d %s(addr=0x%08x) -> error (%s)", num, "" if now else "...", addr, error)
                self._handle_error(error, num)
                raise
            finally:
                if did_lock:
                    self.unlock()

        if now:
            return read_ap_cb()
        else:
            TRACE.debug("read_ap:%06d (addr=0x%08x) -> ...", num, addr)
            return read_ap_cb

    def write_ap_multiple(self, addr, values):
        assert isinstance(addr, int)
        num = self.next_access_number
        did_lock = False

        try:
            did_lock = self._select_ap(addr)
            TRACE.debug("write_ap_multiple:%06d (addr=0x%08x) = (%i values)", num, addr, len(values))
            return self.probe.write_ap_multiple(addr, values)
        except exceptions.TargetError as error:
            self._handle_error(error, num)
            raise
        finally:
            if did_lock:
                self.unlock()

    def read_ap_multiple(self, addr, count=1, now=True):
        assert isinstance(addr, int)
        num = self.next_access_number
        did_lock = False
        
        try:
            did_lock = self._select_ap(addr)
            TRACE.debug("read_ap_multiple:%06d (addr=0x%08x, count=%i)", num, addr, count)
            result_cb = self.probe.read_ap_multiple(addr, count, now=False)
        except exceptions.TargetError as error:
            self._handle_error(error, num)
            raise
        except Exception:
            if did_lock:
                self.unlock()
            raise

        # Need to wrap the deferred callback to convert exceptions.
        def read_ap_multiple_cb():
            try:
                return result_cb()
            except exceptions.TargetError as error:
                TRACE.debug("read_ap_multiple:%06d %s(addr=0x%08x) -> error (%s)", num, "" if now else "...", addr, error)
                self._handle_error(error, num)
                raise
            finally:
                if did_lock:
                    self.unlock()

        if now:
            return read_ap_multiple_cb()
        else:
            return read_ap_multiple_cb

    def _handle_error(self, error, num):
        TRACE.debug("error:%06d %s", num, error)
        # Clear sticky error for fault errors.
        if isinstance(error, exceptions.TransferFaultError):
            self.clear_sticky_err()
        # For timeouts caused by WAIT responses, set DAPABORT to abort the transfer.
        elif isinstance(error, exceptions.TransferTimeoutError):
            # This may put the AP that was aborted into an unpredictable state. Should consider
            # attempting to reset debug logic.
            self.write_reg(DP_ABORT, ABORT_DAPABORT)

    def clear_sticky_err(self):
        self._invalidate_cache()
        mode = self.probe.wire_protocol
        if mode == DebugProbe.Protocol.SWD:
            self.write_reg(DP_ABORT, ABORT_ORUNERRCLR | ABORT_WDERRCLR | ABORT_STKERRCLR | ABORT_STKCMPCLR)
        elif mode == DebugProbe.Protocol.JTAG:
            self.write_reg(DP_CTRL_STAT, CSYSPWRUPREQ | CDBGPWRUPREQ | TRNNORMAL | MASKLANE
                    | CTRLSTAT_STICKYERR | CTRLSTAT_STICKYCMP | CTRLSTAT_STICKYORUN)
        else:
            assert False

class APAccessMemoryInterface(memory_interface.MemoryInterface):
    """! @brief Memory interface for performing simple APACC transactions.
    
    This class allows the caller to generate Debug APB transactions from a DPv3. It simply
    adapts the MemoryInterface to APACC transactions.
    
    By default, it passes memory transaction addresses unmodified to the DP. But an instance can be
    constructed by passing an APAddress object to the constructor that offsets transaction addresses
    so they are relative to the APAddress base.
    
    Only 32-bit transfers are supported.
    """
    
    def __init__(self, dp, ap_address=None):
        """! @brief Constructor.
        
        @param self
        @param dp The DebugPort object.
        @param ap_address Optional instance of APAddress. If provided, all memory transaction
            addresses are offset by the base address of the APAddress.
        """
        self._dp = dp
        self._ap_address = ap_address
        if ap_address is not None:
            self._offset = ap_address.address
        else:
            self._offset = 0
    
    @property
    def dp(self):
        return self._dp
    
    @property
    def short_description(self):
        if self._ap_address is None:
            return "Root Component"
        else:
            return "Root Component ({})".format(self._ap_address)

    def write_memory(self, addr, data, transfer_size=32):
        """! @brief Write a single memory location.
        
        By default the transfer size is a word."""
        if transfer_size != 32:
            raise exceptions.DebugError("unsupported transfer size")
        
        return self._dp.write_ap(self._offset + addr, data)
        
    def read_memory(self, addr, transfer_size=32, now=True):
        """! @brief Read a memory location.
        
        By default, a word will be read."""
        if transfer_size != 32:
            raise exceptions.DebugError("unsupported transfer size")
        
        return self._dp.read_ap(self._offset + addr, now)

    def write_memory_block32(self, addr, data):
        """! @brief Write an aligned block of 32-bit words."""
        addr += self._offset
        for word in data:
            self._dp.write_ap(addr, data)
            addr += 4

    def read_memory_block32(self, addr, size):
        """! @brief Read an aligned block of 32-bit words."""
        addr += self._offset
        result_cbs = [self._dp.read_ap(addr + i * 4, now=False) for i in range(size)]
        result = [cb() for cb in result_cbs]
        return result




# pyOCD debugger
# Copyright (c) 2020 Cypress Semiconductor Corporation
# SPDX-License-Identifier: Apache-2.0
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import logging

from ...family.flash_psoc6 import Flash_PSoC64, PSoC6FlashParams
from ...family.target_psoc6 import PSoC64, CortexM_PSoC64_A2M
from ....core.memory_map import (FlashRegion, MemoryMap)

LOG = logging.getLogger(__name__)


class cy8c64xA(PSoC64):
    from .flash_algos.flash_algo_CY8C64xA import flash_algo as flash_algo_main
    from .flash_algos.flash_algo_CY8C6xxA_WFLASH import flash_algo as flash_algo_work
    from .flash_algos.flash_algo_CY8C6xxA_SMIF_S25FL512S import flash_algo as flash_algo_smif

    MEMORY_MAP = MemoryMap(
        PSoC6FlashParams.defaultRomRegion,
        PSoC6FlashParams.defaultRamRegion,

        FlashRegion(start=0x10000000, length=0x1E0000, blocksize=0x200,
                    is_boot_memory=True,
                    erased_byte_value=0,
                    algo=flash_algo_main,
                    erase_all_weight=PSoC6FlashParams.MFLASH_ERASE_ALL_WEIGHT,
                    erase_sector_weight=PSoC6FlashParams.MFLASH_ERASE_SECTOR_WEIGHT,
                    program_page_weight=PSoC6FlashParams.MFLASH_PROGRAM_PAGE_WEIGHT,
                    flash_class=Flash_PSoC64),

        FlashRegion(start=0x14000000, length=0x8000, blocksize=0x200,
                    is_boot_memory=False,
                    erased_byte_value=0,
                    algo=flash_algo_work,
                    erase_all_weight=PSoC6FlashParams.MFLASH_ERASE_ALL_WEIGHT,
                    erase_sector_weight=PSoC6FlashParams.MFLASH_ERASE_SECTOR_WEIGHT,
                    program_page_weight=PSoC6FlashParams.MFLASH_PROGRAM_PAGE_WEIGHT,
                    flash_class=Flash_PSoC64),

        FlashRegion(start=0x18000000, length=0x4000000, blocksize=0x40000, page_size=0x1000,
                    is_boot_memory=False,
                    is_testable=False,
                    erased_byte_value=0xFF,
                    is_powered_on_boot=False,
                    algo=flash_algo_smif,
                    erase_all_weight=PSoC6FlashParams.SMIF_ERASE_ALL_WEIGHT,
                    erase_sector_weight=PSoC6FlashParams.SMIF_ERASE_SECTOR_WEIGHT,
                    program_page_weight=PSoC6FlashParams.SMIF_PROGRAM_PAGE_WEIGHT,
                    flash_class=Flash_PSoC64),
    )

    def __init__(self, session, ap_num):
        super(cy8c64xA, self).__init__(session, CortexM_PSoC64_A2M, self.MEMORY_MAP, ap_num)


class cy8c64xA_cm0(cy8c64xA):
    def __init__(self, session):
        super(cy8c64xA_cm0, self).__init__(session, 1)


class cy8c64xA_cm4(cy8c64xA):
    def __init__(self, session):
        super(cy8c64xA_cm4, self).__init__(session, 2)


class cy8c64xA_cm4_full_flash(cy8c64xA_cm4):
    from .flash_algos.flash_algo_CY8C64xA import flash_algo as flash_algo_main
    from .flash_algos.flash_algo_CY8C6xxA_WFLASH import flash_algo as flash_algo_work
    from .flash_algos.flash_algo_CY8C6xxA_SMIF_S25FL512S import flash_algo as flash_algo_smif

    MEMORY_MAP = MemoryMap(
        PSoC6FlashParams.defaultRomRegion,
        PSoC6FlashParams.defaultRamRegion,

        FlashRegion(start=0x10000000, length=0x200000, blocksize=0x200,
                    is_boot_memory=True,
                    erased_byte_value=0,
                    algo=flash_algo_main,
                    erase_all_weight=PSoC6FlashParams.MFLASH_ERASE_ALL_WEIGHT,
                    erase_sector_weight=PSoC6FlashParams.MFLASH_ERASE_SECTOR_WEIGHT,
                    program_page_weight=PSoC6FlashParams.MFLASH_PROGRAM_PAGE_WEIGHT,
                    flash_class=Flash_PSoC64),

        FlashRegion(start=0x14000000, length=0x8000, blocksize=0x200,
                    is_boot_memory=False,
                    erased_byte_value=0,
                    algo=flash_algo_work,
                    erase_all_weight=PSoC6FlashParams.MFLASH_ERASE_ALL_WEIGHT,
                    erase_sector_weight=PSoC6FlashParams.MFLASH_ERASE_SECTOR_WEIGHT,
                    program_page_weight=PSoC6FlashParams.MFLASH_PROGRAM_PAGE_WEIGHT,
                    flash_class=Flash_PSoC64),

        FlashRegion(start=0x18000000, length=0x4000000, blocksize=0x40000, page_size=0x1000,
                    is_boot_memory=False,
                    is_testable=False,
                    erased_byte_value=0xFF,
                    is_powered_on_boot=False,
                    algo=flash_algo_smif,
                    erase_all_weight=PSoC6FlashParams.SMIF_ERASE_ALL_WEIGHT,
                    erase_sector_weight=PSoC6FlashParams.SMIF_ERASE_SECTOR_WEIGHT,
                    program_page_weight=PSoC6FlashParams.SMIF_PROGRAM_PAGE_WEIGHT,
                    flash_class=Flash_PSoC64),
    )


class cy8c64xA_cm0_full_flash(cy8c64xA_cm0):
    from .flash_algos.flash_algo_CY8C64xA import flash_algo as flash_algo_main
    from .flash_algos.flash_algo_CY8C6xxA_WFLASH import flash_algo as flash_algo_work
    from .flash_algos.flash_algo_CY8C6xxA_SMIF_S25FL512S import flash_algo as flash_algo_smif

    MEMORY_MAP = MemoryMap(
        PSoC6FlashParams.defaultRomRegion,
        PSoC6FlashParams.defaultRamRegion,

        FlashRegion(start=0x10000000, length=0x200000, blocksize=0x200,
                    is_boot_memory=True,
                    erased_byte_value=0,
                    algo=flash_algo_main,
                    erase_all_weight=PSoC6FlashParams.MFLASH_ERASE_ALL_WEIGHT,
                    erase_sector_weight=PSoC6FlashParams.MFLASH_ERASE_SECTOR_WEIGHT,
                    program_page_weight=PSoC6FlashParams.MFLASH_PROGRAM_PAGE_WEIGHT,
                    flash_class=Flash_PSoC64),

        FlashRegion(start=0x14000000, length=0x8000, blocksize=0x200,
                    is_boot_memory=False,
                    erased_byte_value=0,
                    algo=flash_algo_work,
                    erase_all_weight=PSoC6FlashParams.MFLASH_ERASE_ALL_WEIGHT,
                    erase_sector_weight=PSoC6FlashParams.MFLASH_ERASE_SECTOR_WEIGHT,
                    program_page_weight=PSoC6FlashParams.MFLASH_PROGRAM_PAGE_WEIGHT,
                    flash_class=Flash_PSoC64),

        FlashRegion(start=0x18000000, length=0x4000000, blocksize=0x40000, page_size=0x1000,
                    is_boot_memory=False,
                    is_testable=False,
                    erased_byte_value=0xFF,
                    is_powered_on_boot=False,
                    algo=flash_algo_smif,
                    erase_all_weight=PSoC6FlashParams.SMIF_ERASE_ALL_WEIGHT,
                    erase_sector_weight=PSoC6FlashParams.SMIF_ERASE_SECTOR_WEIGHT,
                    program_page_weight=PSoC6FlashParams.SMIF_PROGRAM_PAGE_WEIGHT,
                    flash_class=Flash_PSoC64),
    )



# pyOCD debugger
# Copyright (c) 2017-2019 Arm Limited
# SPDX-License-Identifier: Apache-2.0
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from time import (time, sleep)
import pytest

from pyocd.utility.timeout import Timeout

class TestTimeout:
    def test_no_timeout(self):
        with Timeout(0.05) as to:
            cnt = 0
            while to.check():
                sleep(0.01)
                cnt += 1
                if cnt == 2:
                    break
            else:
                assert False
        assert not to.did_time_out

    def test_timeout_a(self):
        s = time()
        with Timeout(0.05) as to:
            while to.check():
                sleep(0.01)
        assert to.did_time_out
        assert (time() - s) >= 0.05
    
    def test_timeout_b(self):
        timedout = False
        s = time()
        with Timeout(0.05) as to:
            cnt = 0
            while cnt < 10:
                if to.did_time_out:
                    timedout = True
                sleep(0.02)
                cnt += 1
        assert timedout
        assert to.did_time_out
        assert (time() - s) >= 0.05
    
    def test_timeout_c(self):
        timedout = False
        with Timeout(0.05) as to:
            cnt = 0
            while cnt < 10:
                if to.did_time_out:
                    timedout = True
                cnt += 1
        assert not timedout
        assert not to.did_time_out
    
    def test_timeout_reset(self):
        cnt = 0
        cnta = 0
        with Timeout(0.05) as to:
            cnta = 0
            while cnta < 3:
                cnt = 0
                while to.check():
                    sleep(0.01)
                    if cnta > 1:
                        break
                    cnt += 1
                else:
                    assert to.did_time_out
                    to.clear()
                    to.start()
                cnta += 1
        assert cnta == 3 and cnt == 0
        assert not to.did_time_out




# pyOCD debugger
# Copyright (c) 2015-2020 Arm Limited
# Copyright (c) 2021 Chris Reed
# SPDX-License-Identifier: Apache-2.0
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import logging
from inspect import getfullargspec
from pathlib import PurePath

from ..core.target import Target
from ..core.memory_map import (MemoryType, RamRegion, DeviceRegion, MemoryMap)
from ..core.soc_target import SoCTarget
from ..core import exceptions
from . import (dap, discovery)
from ..debug.svd.loader import SVDLoader
from ..utility.sequencer import CallSequence
from ..target.pack.flash_algo import PackFlashAlgo

LOG = logging.getLogger(__name__)

class CoreSightTarget(SoCTarget):
    """! @brief Represents an SoC that uses CoreSight debug infrastructure.
    
    This class adds Arm CoreSight-specific discovery and initialization code to SoCTarget.
    """
    
    def __init__(self, session, memory_map=None):
        # Supply a default memory map.
        if (memory_map is None) or (memory_map.region_count == 0):
            memory_map = self._create_default_cortex_m_memory_map()
            LOG.debug("Using default Cortex-M memory map (no memory map supplied)")
        
        super(CoreSightTarget, self).__init__(session, memory_map)
        self.dp = dap.DebugPort(session.probe, self)
        self._svd_load_thread = None
        self._irq_table = None
        self._discoverer = None

    @property
    def aps(self):
        return self.dp.aps

    @property
    def svd_device(self):
        """! @brief Waits for SVD file to complete loading before returning."""
        if not self._svd_device and self._svd_load_thread:
            LOG.debug("Waiting for SVD load to complete")
            self._svd_device = self._svd_load_thread.device
        return self._svd_device

    def _create_default_cortex_m_memory_map(self):
        """! @brief Create a MemoryMap for the Cortex-M system address map."""
        return MemoryMap(
                RamRegion(name="Code",          start=0x00000000, length=0x20000000, access='rwx'),
                RamRegion(name="SRAM",          start=0x20000000, length=0x20000000, access='rwx'),
                DeviceRegion(name="Peripheral", start=0x40000000, length=0x20000000, access='rw'),
                RamRegion(name="RAM1",          start=0x60000000, length=0x20000000, access='rwx'),
                RamRegion(name="RAM2",          start=0x80000000, length=0x20000000, access='rwx'),
                DeviceRegion(name="Device1",    start=0xA0000000, length=0x20000000, access='rw'),
                DeviceRegion(name="Device2",    start=0xC0000000, length=0x20000000, access='rw'),
                DeviceRegion(name="PPB",        start=0xE0000000, length=0x20000000, access='rw'),
                )

    def load_svd(self):
        def svd_load_completed_cb(svdDevice):
            self._svd_device = svdDevice
            self._svd_load_thread = None

        if not self._svd_device and self._svd_location:
            # Spawn thread to load SVD in background.
            self._svd_load_thread = SVDLoader(self._svd_location, svd_load_completed_cb)
            self._svd_load_thread.load()

    def create_init_sequence(self):
        seq = CallSequence(
            ('load_svd',            self.load_svd),
            ('pre_connect',         self.pre_connect),
            ('dp_init',             self.dp.create_connect_sequence),
            ('create_discoverer',   self.create_discoverer),
            ('discovery',           lambda : self._discoverer.discover()),
            ('check_for_cores',     self.check_for_cores),
            ('halt_on_connect',     self.perform_halt_on_connect),
            ('post_connect',        self.post_connect),
            ('post_connect_hook',   self.post_connect_hook),
            ('create_flash',        self.create_flash),
            ('notify',              lambda : self.session.notify(Target.Event.POST_CONNECT, self))
            )
        
        return seq

    def disconnect(self, resume=True):
        """! @brief Disconnect from the target.
        
        Same as SoCTarget.disconnect(), except that it asks the DebugPort to power down.
        """
        self.session.notify(Target.Event.PRE_DISCONNECT, self)
        self.call_delegate('will_disconnect', target=self, resume=resume)
        for core in self.cores.values():
            core.disconnect(resume)
        self.dp.disconnect()
        self.call_delegate('did_disconnect', target=self, resume=resume)
            
    def create_discoverer(self):
        """! @brief Init task to create the discovery object.
        
        Instantiates the appropriate @ref pyocd.coresight.discovery.CoreSightDiscovery
        CoreSightDiscovery subclass for the target's ADI version.
        """
        self._discoverer = discovery.ADI_DISCOVERY_CLASS_MAP[self.dp.adi_version](self)

    def pre_connect(self):
        """! @brief Handle some of the connect modes.
        
        This init task performs a connect pre-reset or asserts reset if the connect mode is
        under-reset.
        """
        mode = self.session.options.get('connect_mode')
        if mode == 'pre-reset':
            LOG.info("Performing connect pre-reset")
            self.dp.reset()
        elif mode == 'under-reset':
            LOG.info("Asserting reset prior to connect")
            self.dp.assert_reset(True)
    
    def perform_halt_on_connect(self):
        """! @brief Halt cores.
        
        This init task performs a connect pre-reset or asserts reset if the connect mode is
        under-reset.
        """
        mode = self.session.options.get('connect_mode')
        if mode != 'attach':
            if mode == 'under-reset':
                LOG.debug("Setting reset catch")
            # Apply to all cores.
            for core in self.cores.values():
                try:
                    if mode == 'under-reset':
                        core.set_reset_catch()
                    else:
                        core.halt()
                except exceptions.Error as err:
                    LOG.warning("Could not halt core #%d: %s", core.core_number, err,
                        exc_info=self.session.log_tracebacks)
    
    def post_connect(self):
        """! @brief Handle cleaning up some of the connect modes.
        
        This init task de-asserts reset if the connect mode is under-reset.
        """
        mode = self.session.options.get('connect_mode')
        if mode == 'under-reset':
            LOG.info("Deasserting reset post connect")
            self.dp.assert_reset(False)

            LOG.debug("Clearing reset catch")
            # Apply to all cores.
            for core in self.cores.values():
                try:
                    core.clear_reset_catch()
                except exceptions.Error as err:
                    LOG.warning("Could not halt core #%d: %s", core.core_number, err,
                        exc_info=self.session.log_tracebacks)
    
    def create_flash(self):
        """! @brief Instantiates flash objects for memory regions.
        
        This init task iterates over flash memory regions and for each one creates the Flash
        instance. It uses the flash_algo and flash_class properties of the region to know how
        to construct the flash object.
        """
        for region in self.memory_map.iter_matching_regions(type=MemoryType.FLASH):
            # If the region doesn't have an algo dict but does have an FLM file, try to load
            # the FLM and create the algo dict.
            if (region.algo is None) and (region.flm is not None):
                if isinstance(region.flm, (str, PurePath)):
                    flm_path = self.session.find_user_file(None, [region.flm])
                    if flm_path is not None:
                        LOG.info("creating flash algo from: %s", flm_path)
                        pack_algo = PackFlashAlgo(flm_path)
                    else:
                        LOG.warning("Failed to find FLM file: %s", region.flm)
                        break
                elif isinstance(region.flm, PackFlashAlgo):
                    pack_algo = region.flm
                else:
                    LOG.warning("flash region flm attribute is unexpected type")
                    break

                # Create the algo dict from the FLM.
                if self.session.options.get("debug.log_flm_info"):
                    LOG.debug("Flash algo info: %s", pack_algo.flash_info)
                page_size = pack_algo.page_size
                if page_size <= 32:
                    page_size = min(s[1] for s in pack_algo.sector_sizes)
                algo = pack_algo.get_pyocd_flash_algo(
                        page_size,
                        self.memory_map.get_default_region_of_type(MemoryType.RAM))
                
                # If we got a valid algo from the FLM, set it on the region. This will then
                # be used below.
                if algo is not None:
                    region.algo = algo
            
            # If the constructor of the region's flash class takes the flash_algo arg, then we
            # need the region to have a flash algo dict to pass to it. Otherwise we assume the
            # algo is built-in.
            klass = region.flash_class
            argspec = getfullargspec(klass.__init__)
            if 'flash_algo' in argspec.args:
                if region.algo is not None:
                    obj = klass(self, region.algo)
                else:
                    LOG.warning("flash region '%s' has no flash algo" % region.name)
                    continue
            else:
                obj = klass(self)
            
            # Set the region in the flash instance.
            obj.region = region
            
            # Store the flash object back into the memory region.
            region.flash = obj

    def check_for_cores(self):
        """! @brief Init task: verify that at least one core was discovered."""
        if not len(self.cores):
            # Allow the user to override the exception to enable uses like chip bringup.
            if self.session.options.get('allow_no_cores'):
                LOG.error("No cores were discovered!")
            else:
                raise exceptions.DebugError("No cores were discovered!")
    @property
    def irq_table(self):
        if (self._irq_table is None):
            if (self.svd_device is not None) and (self.svd_device.peripherals is not None):
                peripherals = [
                        p for p in self.svd_device.peripherals
                        if p.interrupts is not None
                        ]
                self._irq_table = {
                        i.value : i.name
                        for p in peripherals
                            for i in p.interrupts
                        }
            else:
                self._irq_table = {}
        return self._irq_table

    # Override this method from SoCTarget so we can use the DP for hardware resets when there isn't a
    # valid core (instead of the probe), so reset notifications will be sent. We can't use the DP in
    # SoCTarget because it is only created by this class.
    def reset(self, reset_type=None):
        # Use the DP to reset if there is not a core.
        if (self.selected_core is None) and (self.dp is not None):
            self.dp.reset()
        else:
            super().reset(reset_type)
    
        



# pyOCD debugger
# Copyright (c) 2018-2019 Arm Limited
# Copyright (c) 2021 Chris Reed
# SPDX-License-Identifier: Apache-2.0
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import pytest
import six

from pyocd.utility.sequencer import CallSequence

class TestCallSequence:
    def test_empty(self):
        cs = CallSequence()
        assert cs.count == 0
    
    def test_a(self):
        results = []
        cs = CallSequence(
                ('a', lambda : results.append('a ran')),
                ('b', lambda : results.append('b ran')),
                )
        assert cs.count == 2
        cs.invoke()
        assert results == ['a ran', 'b ran']

    def test_append_1(self):
        results = []
        cs = CallSequence(
                ('a', lambda : results.append('a ran')),
                )
        assert cs.count == 1

        cs.append(        
                ('b', lambda : results.append('b ran')),
                )
        assert cs.count == 2
        
        cs.invoke()
        assert results == ['a ran', 'b ran']

    def test_append_2(self):
        results = []
        cs = CallSequence(
                ('a', lambda : results.append('a ran')),
                )
        assert cs.count == 1

        cs.append(        
                ('b', lambda : results.append('b ran')),
                ('c', lambda : results.append('c ran')),
                )
        assert cs.count == 3
        
        cs.invoke()
        assert results == ['a ran', 'b ran', 'c ran']

    def test_remove_1(self):
        results = []
        cs = CallSequence(
                ('a', lambda : results.append('a ran')),
                ('b', lambda : results.append('b ran')),
                )
        assert cs.count == 2
        
        cs.remove_task('b')
        assert cs.count == 1
        
        cs.invoke()
        assert results == ['a ran']

    def test_callable(self):
        results = []
        cs = CallSequence(
                ('a', lambda : results.append('a ran')),
                ('b', lambda : results.append('b ran')),
                )
        assert cs.count == 2
        cs()
        assert results == ['a ran', 'b ran']

    def test_nested(self):
        results = []
        cs = CallSequence(
                ('a', lambda : results.append('a ran')),
                ('b', lambda : results.append('b ran')),
                )
        assert cs.count == 2
        cs2 = CallSequence(
                ('c', cs),
                )
        assert cs2.count == 1
        cs2.invoke()
        assert results == ['a ran', 'b ran']

    def test_clear(self):
        results = []
        cs = CallSequence(
                ('a', lambda : results.append('a ran')),
                ('b', lambda : results.append('b ran')),
                )
        assert cs.count == 2
        cs.clear()
        assert cs.count == 0

    def test_iter(self):
        results = []
        def task_a():
            results.append('a ran')
        def task_b():
            results.append('b ran')
        cs = CallSequence(
                ('a', task_a),
                ('b', task_b),
                )
        assert cs.count == 2
        it = iter(cs)
        print("it=",repr(it),dir(it))
        assert six.next(it) == ('a', task_a)
        assert six.next(it) == ('b', task_b)
        with pytest.raises(StopIteration):
            six.next(it)

    def test_get(self):
        results = []
        def task_a():
            results.append('a ran')
        cs = CallSequence(
                ('a', task_a),
                )
        assert cs.count == 1
        assert cs.get_task('a') == task_a
        with pytest.raises(KeyError):
            cs.get_task('foo')

    def test_has(self):
        results = []
        cs = CallSequence(
                ('a', lambda : results.append('a ran')),
                )
        assert cs.count == 1
        assert cs.has_task('a')
        assert not cs.has_task('foo')

    def test_replace(self):
        results = []
        cs = CallSequence(
                ('a', lambda : results.append('a ran')),
                ('b', lambda : results.append('b ran')),
                )
        assert cs.count == 2
        cs.replace_task('b', lambda : results.append('wheee'))
        cs()
        assert results == ['a ran', 'wheee']

    def test_wrap(self):
        results = []
        def task_b():
            results.append('b ran')
            return "task b result"
        cs = CallSequence(
                ('a', lambda : results.append('a ran')),
                ('b', task_b),
                )
        assert cs.count == 2
        def wrapper(t):
            assert t == "task b result"
            results.append('wrapper ran')
        cs.wrap_task('b', wrapper)
        cs()
        assert results == ['a ran', 'b ran', 'wrapper ran']

    def test_returned_seq(self):
        results = []
        def task_b():
            results.append('b ran')
            cs2 = CallSequence(
                    ('x', lambda : results.append('x ran')),
                    ('y', lambda : results.append('y ran')),
                    )
            assert cs2.count == 2
            return cs2
        cs = CallSequence(
                ('a', lambda : results.append('a ran')),
                ('b', task_b),
                ('c', lambda : results.append('c ran')),
                )
        assert cs.count == 3
        cs()
        assert results == ['a ran', 'b ran', 'x ran', 'y ran', 'c ran']

    def test_insert_before_1(self):
        results = []
        cs = CallSequence(
                ('a', lambda : results.append('a ran')),
                ('b', lambda : results.append('b ran')),
                )
        assert cs.count == 2
        cs.insert_before('b', ('c', lambda : results.append('c ran')))
        assert cs.count == 3
        cs()
        assert results == ['a ran', 'c ran', 'b ran']

    def test_insert_before_2(self):
        results = []
        cs = CallSequence(
                ('a', lambda : results.append('a ran')),
                ('b', lambda : results.append('b ran')),
                )
        assert cs.count == 2
        cs.insert_before('a', ('c', lambda : results.append('c ran')))
        assert cs.count == 3
        cs()
        assert results == ['c ran', 'a ran', 'b ran']

    def test_insert_before_3(self):
        results = []
        cs = CallSequence(
                ('a', lambda : results.append('a ran')),
                ('b', lambda : results.append('b ran')),
                )
        assert cs.count == 2
        cs.insert_before('a', ('c', lambda : results.append('c ran')),
                              ('d', lambda : results.append('d ran')))
        assert cs.count == 4
        cs()
        assert results == ['c ran', 'd ran', 'a ran', 'b ran']

    def test_insert_before_4(self):
        results = []
        cs = CallSequence()
        with pytest.raises(KeyError):
            cs.insert_before('z', ('c', lambda : results.append('c ran')))

    def test_insert_after_1(self):
        results = []
        cs = CallSequence(
                ('a', lambda : results.append('a ran')),
                ('b', lambda : results.append('b ran')),
                )
        assert cs.count == 2
        cs.insert_after('b', ('c', lambda : results.append('c ran')))
        assert cs.count == 3
        cs()
        assert results == ['a ran', 'b ran', 'c ran']

    def test_insert_after_2(self):
        results = []
        cs = CallSequence(
                ('a', lambda : results.append('a ran')),
                ('b', lambda : results.append('b ran')),
                )
        assert cs.count == 2
        cs.insert_after('a', ('c', lambda : results.append('c ran')))
        assert cs.count == 3
        cs()
        assert results == ['a ran', 'c ran', 'b ran']

    def test_insert_after_3(self):
        results = []
        cs = CallSequence(
                ('a', lambda : results.append('a ran')),
                ('b', lambda : results.append('b ran')),
                )
        assert cs.count == 2
        cs.insert_after('a', ('c', lambda : results.append('c ran')),
                             ('d', lambda : results.append('d ran')))
        assert cs.count == 4
        cs()
        assert results == ['a ran', 'c ran', 'd ran', 'b ran']

    def test_insert_after_4(self):
        results = []
        cs = CallSequence()
        with pytest.raises(KeyError):
            cs.insert_after('z', ('c', lambda : results.append('c ran')))





# pyOCD debugger
# Copyright (c) 2019-2020 Arm Limited
# SPDX-License-Identifier: Apache-2.0
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import logging

from ..core import exceptions
from .ap import (APv1Address, APv2Address, AccessPort)
from .dap import (ADIVersion, APAccessMemoryInterface)
from .rom_table import (CoreSightComponentID, ROMTable)
from . import (cortex_m, cortex_m_v8m)
from ..utility.sequencer import CallSequence

LOG = logging.getLogger(__name__)

class CoreSightDiscovery(object):
    """! @brief Base class for discovering CoreSight components in a target."""

    def __init__(self, target):
        """! @brief Constructor."""
        self._target = target
    
    @property
    def target(self):
        return self._target
    
    @property
    def dp(self):
        return self.target.dp
    
    @property
    def session(self):
        return self.target.session

    def discover(self):
        """! @brief Init task for component discovery.
        @return CallSequence for the discovery process.
        """
        raise NotImplementedError()

    def _create_component(self, cmpid):
        try:
            LOG.debug("Creating %s component", cmpid.name)
            cmp = cmpid.factory(cmpid.ap, cmpid, cmpid.address)
            cmp.init()
        except exceptions.Error as err:
            LOG.error("Error attempting to create component %s: %s", cmpid.name, err,
                    exc_info=self.session.log_tracebacks)

    def _create_cores(self):
        self._apply_to_all_components(self._create_component,
            filter=lambda c: c.factory in (cortex_m.CortexM.factory, cortex_m_v8m.CortexM_v8M.factory))

    def _create_components(self):
        self._apply_to_all_components(self._create_component,
            filter=lambda c: c.factory is not None
                and c.factory not in (cortex_m.CortexM.factory, cortex_m_v8m.CortexM_v8M.factory))
    
    def _apply_to_all_components(self, action, filter=None):
        # Iterate over every top-level ROM table.
        for ap in [x for x in self.dp.aps.values() if x.rom_table]:
            ap.rom_table.for_each(action, filter)

class ADIv5Discovery(CoreSightDiscovery):
    """! @brief Component discovery process for ADIv5.
    
    Component discovery for ADIv5 proceeds as follows. Each of the steps is labeled with the name
    of the init task for that step.
    
    1. `find_aps`: Perform an AP scan. Probe each AP at APSEL=0..255. By default the scan stops on
        the first invalid APSEL, as determined by testing the IDR value (0 is invalid). This can be
        overridden by a session option.
    2. `create_aps`: Create all APs and add them to the DP.
    3. `find_components`: For each AP, read the associated ROM table(s) and identify CoreSight
        components.
    4. `create_cores`: Create any discovered core (CPU) components. The cores are created first to
        ensure that other components have a core to which they may be connected.
    5. `create_components`: Create remaining discovered components.
    """

    ## APSEL is 8-bit, thus there are a maximum of 256 APs.
    MAX_APSEL = 255

    def discover(self):
        return CallSequence(
            ('find_aps',            self._find_aps),
            ('create_aps',          self._create_aps),
            ('find_components',     self._find_components),
            ('create_cores',        self._create_cores),
            ('create_components',   self._create_components),
            )

    def _find_aps(self):
        """! @brief Find valid APs using the ADIv5 method.
        
        Scans for valid APs starting at APSEL=0. The default behaviour is to stop after reading
        0 for the AP's IDR twice in succession. If the `scan_all_aps` session option is set to True,
        then the scan will instead probe every APSEL from 0-255.
        
        If there is already a list of valid APs defined for the @ref pyocd.coresight.dap.DebugPort
        DebugPort (the `valid_aps` attribute), then scanning is not performed. This is to allow a
        predetermined list of valid APSELs to be used in place of a scan. A few MCUs will lock up
        when accessing invalid APs. On those devices, scanning with the method used here cannot be
        done.
        """
        # Don't perform the AP scan if there is already a list of valid APs. This is to allow
        # skipping the AP scan by providing a predetermined list of valid APSELs.
        if self.dp.valid_aps is not None:
            return
        
        ap_list = []
        apsel = 0
        invalid_count = 0
        while apsel < self.MAX_APSEL:
            try:
                isValid = AccessPort.probe(self.dp, apsel)
                if isValid:
                    ap_list.append(apsel)
                    invalid_count = 0
                elif not self.session.options.get('scan_all_aps'):
                    invalid_count += 1
                    if invalid_count == self.session.options.get('adi.v5.max_invalid_ap_count'):
                        break
            except exceptions.Error as e:
                LOG.error("Exception while probing AP#%d: %s", apsel, e,
                    exc_info=self.session.log_tracebacks)
                break
            apsel += 1
        
        # Update the AP list once we know it's complete.
        self.dp.valid_aps = ap_list

    def _create_aps(self):
        """! @brief Init task that returns a call sequence to create APs.
        
        For each AP in the #valid_aps list, an AccessPort object is created. The new objects
        are added to the #aps dict, keyed by their AP number.
        """
        seq = CallSequence()
        for apsel in self.dp.valid_aps:
            seq.append(
                ('create_ap.{}'.format(apsel), lambda apsel=apsel: self._create_1_ap(apsel))
                )
        return seq
    
    def _create_1_ap(self, apsel):
        """! @brief Init task to create a single AP object."""
        try:
            ap_address = APv1Address(apsel)
            ap = AccessPort.create(self.dp, ap_address)
            self.dp.aps[ap_address] = ap
        except exceptions.Error as e:
            LOG.error("Exception reading AP#%d IDR: %s", apsel, e,
                exc_info=self.session.log_tracebacks)
    
    def _find_components(self):
        """! @brief Init task that generates a call sequence to ask each AP to find its components."""
        seq = CallSequence()
        for ap in [x for x in self.dp.aps.values() if x.has_rom_table]:
            seq.append(
                ('init_ap.{}'.format(ap.address.apsel), ap.find_components)
                )
        return seq

class ADIv6Discovery(CoreSightDiscovery):
    """! @brief Component discovery process for ADIv6.
    
    The process for discovering components in ADIv6 proceeds as follows. Each of the steps is
    labeled with the name of the init task for that step.
    
    1. `find_root_components`: Examine the component pointed to by the DP BASEPTR register(s). If
        it's a ROM table, read it and examine components pointed to by the entries. This creates the
        AP instances.
    2. `find_components`: For each AP, read the associated ROM table(s) and identify CoreSight
        components.
    3. `create_cores`: Create any discovered core (CPU) components. The cores are created first to
        ensure that other components have a core to which they may be connected.
    4. `create_components`: Create remaining discovered components.
    
    Note that nested APs are not supported.
    """

    def __init__(self, target):
        """! @brief Constructor."""
        super(ADIv6Discovery, self).__init__(target)
        self._top_rom_table = None

    def discover(self):
        return CallSequence(
            ('find_root_components',    self._find_root_components),
            ('find_components',         self._find_components_on_aps),
            ('create_cores',            self._create_cores),
            ('create_components',       self._create_components),
            )

    def _find_root_components(self):
        """! @brief Read top-level ROM table pointed to by the DP."""
        # There's not much we can do if we don't have a base address.
        if self.dp.base_address is None:
            return
        
        # Create a temporary memory interface.
        mem_interface = self.dp.apacc_memory_interface
        
        # Examine the base component.
        cmpid = CoreSightComponentID(None, mem_interface, self.dp.base_address)
        cmpid.read_id_registers()
        LOG.debug("Base component: %s", cmpid)
        
        if cmpid.is_rom_table:
            self._top_rom_table = ROMTable.create(mem_interface, cmpid)
            self._top_rom_table.init()
            
            # Create components defined in the DP ROM table.
            self._top_rom_table.for_each(self._create_1_ap,
                    filter=lambda c: c.factory == AccessPort.create)
            
            # Create non-AP components in the DP ROM table.
            self._top_rom_table.for_each(self._create_root_component,
                    filter=lambda c: (c.factory is not None) and (c.factory != AccessPort.create))
        elif cmpid.factory == AccessPort.create:
            self._create_1_ap(cmpid)
        else:
            self._create_root_component(cmpid)
    
    def _create_1_ap(self, cmpid):
        """! @brief Init task to create a single AP object."""
        try:
            ap_address = APv2Address(cmpid.address)
            ap = AccessPort.create(self.dp, ap_address, cmpid=cmpid)
            self.dp.aps[ap_address] = ap
        except exceptions.Error as e:
            LOG.error("Exception reading AP@0x%08x IDR: %s", cmpid.address, e,
                    exc_info=self.session.log_tracebacks)
    
    def _create_root_component(self, cmpid):
        """! @brief Init task to create a component attached directly to the DP.
        
        The newly created component is attached directly to the target instance (i.e.,
        CoreSightTarget or subclass) in the object graph.
        """
        try:
            # Create a memory interface for this component.
            ap_address = APv2Address(cmpid.address)
            memif = APAccessMemoryInterface(self.dp, ap_address)
            
            # Instantiate the component and attach to the target.
            component = cmpid.factory(memif, cmpid, cmpid.address)
            self.target.add_child(component)
            component.init()
        except exceptions.Error as e:
            LOG.error("Exception creating root component at address 0x%08x: %s", cmpid.address, e,
                    exc_info=self.session.log_tracebacks)
    
    def _find_components_on_aps(self):
        """! @brief Init task that generates a call sequence to ask each AP to find its components."""
        seq = CallSequence()
        for ap in [x for x in self.dp.aps.values() if x.has_rom_table]:
            seq.append(
                ('init_ap.{}'.format(str(ap.address)), ap.find_components)
                )
        return seq

## Map from ADI version to the discovery class.
ADI_DISCOVERY_CLASS_MAP = {
        ADIVersion.ADIv5: ADIv5Discovery,
        ADIVersion.ADIv6: ADIv6Discovery,
    }



# pyOCD debugger
# Copyright (c) 2006-2015 Arm Limited
# SPDX-License-Identifier: Apache-2.0
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import logging
from ...core.target import Target
from ...coresight.coresight_target import CoreSightTarget
from ...core.memory_map import (FlashRegion, RamRegion, MemoryMap, DefaultFlashWeights)
from ...debug.svd.loader import SVDFile

LOG = logging.getLogger(__name__)

LARGE_PAGE_START_ADDR = 0x10000
SMALL_PAGE_SIZE = 0x1000
LARGE_PAGE_SIZE = 0x8000
LARGE_TO_SMALL_RATIO = LARGE_PAGE_SIZE / SMALL_PAGE_SIZE
LARGE_ERASE_SECTOR_WEIGHT = DefaultFlashWeights.ERASE_SECTOR_WEIGHT * LARGE_TO_SMALL_RATIO
LARGE_PROGRAM_PAGE_WEIGHT = DefaultFlashWeights.PROGRAM_PAGE_WEIGHT * LARGE_TO_SMALL_RATIO
WRITE_SIZE = 1024

FLASH_ALGO = { 'load_address' : 0x10000000,
               'instructions' : [
                                0xe00abe00, 0x062d780d, 0x24084068, 0xd3000040, 0x1e644058, 0x1c49d1fa, 0x2a001e52, 0x4770d1f2,
                                0x7803e005, 0x42931c40, 0x2001d001, 0x1e494770, 0x2000d2f7, 0x00004770, 0x28100b00, 0x210ed302,
                                0x00d0eb01, 0x486c4770, 0x7801b510, 0x0102f021, 0x22aa7001, 0x23557302, 0x78017303, 0x0101f021,
                                0x73027001, 0xf8d07303, 0xf0411120, 0xf8c00120, 0xf1a01120, 0xf8d00080, 0x064911a0, 0xf100d5fb,
                                0x24010080, 0x408cf880, 0x0113f04f, 0x73026041, 0x78017303, 0x0101f041, 0x73027001, 0xf1a07303,
                                0xf8d00080, 0x01491088, 0xf100d5fb, 0x2107006d, 0x1097f880, 0x0109f04f, 0x109bf880, 0xf0417cc1,
                                0x74c10102, 0x77c377c2, 0x4c2df800, 0xf64e494b, 0x44492060, 0xf04f6008, 0xbd100000, 0x47702000,
                                0x41f0e92d, 0x20324c46, 0x2500444c, 0xe884271d, 0xf10400a1, 0x4e430114, 0x46204688, 0x696047b0,
                                0x2034b960, 0x00a1e884, 0x4641483c, 0x68004448, 0x462060e0, 0x696047b0, 0xd0002800, 0xe8bd2001,
                                0xe92d81f0, 0xf7ff41f0, 0x4d35ff87, 0x444d4604, 0xe9c52032, 0xf1050400, 0x4e320114, 0x4628460f,
                                0x47b060ac, 0xb9686968, 0xe9c52034, 0x482b0400, 0x444860ac, 0x68004639, 0x462860e8, 0x696847b0,
                                0xd0dc2800, 0xe7da2001, 0x41f0e92d, 0x46140006, 0x4925d11d, 0x02fcf8d4, 0xd03a4288, 0x42884923,
                                0x4923d037, 0xd0344288, 0x4131ea4f, 0xd0304288, 0x0100e9d4, 0xe9d44408, 0x44111202, 0x69214408,
                                0x69614408, 0x69a14408, 0x42404408, 0x463061e0, 0xff42f7ff, 0x21324d12, 0x4f12444d, 0x1000e9c5,
                                0x0114f105, 0x468860a8, 0x47b84628, 0xb9806968, 0xe9c52033, 0xf44f0600, 0xe9c56080, 0x48074002,
                                0x44484641, 0x61286800, 0x47b84628, 0x28006968, 0x2001d095, 0x0000e793, 0x400fc080, 0x00000004,
                                0x00000008, 0x1fff1ff1, 0x4e697370, 0x12345678, 0x87654321, 0x00000000, 0x00000000
                                ],
               'pc_init' : 0x10000047,
               'pc_eraseAll' : 0x100000e1,
               'pc_erase_sector' : 0x10000123,
               'pc_program_page' : 0x10000169,
               'begin_data' : 0x2007c000,       # Analyzer uses a max of 120 B data (30 pages * 4 bytes / page)
               # Double buffering is not supported since there is not enough ram
               'page_buffers': [0x2007c000, 0x2007c400],
               'begin_stack' : 0x10001000,
               'static_base' : 0x10000214,
               'min_program_length' : 256,
               'analyzer_supported' : True,
               'analyzer_address' : 0x10002000  # Analyzer 0x10002000..0x10002600
              }

class LPC1768(CoreSightTarget):

    VENDOR = "NXP"
    
    MEMORY_MAP = MemoryMap(
        FlashRegion(    start=0,           length=0x10000,      is_boot_memory=True,
                                                                blocksize=0x1000,
                                                                page_size=0x400,
                                                                phrase_size=256,
                                                                algo=FLASH_ALGO),
        FlashRegion(    start=0x10000,     length=0x70000,      blocksize=0x8000,
                                                                page_size=0x400,
                                                                phrase_size=256,
                                                                algo=FLASH_ALGO,
                                                                erase_sector_weight=LARGE_ERASE_SECTOR_WEIGHT,
                                                                program_page_weight=LARGE_PROGRAM_PAGE_WEIGHT),
        RamRegion(      start=0x10000000,  length=0x8000),
        RamRegion(      start=0x2007C000,  length=0x8000)
        )

    def __init__(self, session):
        super(LPC1768, self).__init__(session, self.MEMORY_MAP)
        self._svd_location = SVDFile.from_builtin("LPC176x5x_v0.2.svd")
        self._saved_vc = False
        self._reset_handler = 0
        self._had_reset_handler_bp = False

    def add_core(self, core):
        super(LPC1768, self).add_core(core)
        core.delegate = self
    
    def map_flash(self):
        self.write32(0x400FC040, 1)

    def set_reset_catch(self, core, reset_type=None):
        LOG.debug("LPC1768 set reset catch")

        # Clear reset vector catch and remember whether it was set.
        self._saved_vc = self.get_vector_catch()
        self.set_vector_catch(self._saved_vc & ~Target.VectorCatch.CORE_RESET)
        
        # Map flash to 0.
        self.map_flash()
        
        # Set breakpoint on user reset handler.
        self._reset_handler = self.read32(0x4)
        if self._reset_handler < 0x80000:
            self._had_reset_handler_bp = (self.get_breakpoint_type(self._reset_handler) is not None)
            self.set_breakpoint(self._reset_handler)

    def clear_reset_catch(self, core, reset_type=None):
        LOG.debug("LPC1768 clear reset catch")

        # Clear breakpoint if it wasn't previously set.
        if not self._had_reset_handler_bp:
            self.remove_breakpoint(self._reset_handler)
        
        # Restore vector catch.
        self.set_vector_catch(self._saved_vc)



# pyOCD debugger
# Copyright (c) 2015-2020 Arm Limited
# Copyright (c) 2021 Chris Reed
# SPDX-License-Identifier: Apache-2.0
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import logging
import os
import traceback

from ..core.helpers import ConnectHelper
from ..core import (exceptions, session)
from ..probe.shared_probe_proxy import SharedDebugProbeProxy
from ..utility.cmdline import convert_session_options
from ..commands.repl import (PyocdRepl, ToolExitException)
from ..commands.execution_context import CommandExecutionContext

LOG = logging.getLogger(__name__)

## Default SWD clock in Hz.
DEFAULT_CLOCK_FREQ_HZ = 1000000

class PyOCDCommander(object):
    """! @brief Manages the commander interface.
    
    Responsible for connecting the execution context, REPL, and commands, and handles connection.
    
    Exit codes:
    - 0 = no errors
    - 1 = command error
    - 2 = transfer error
    - 3 = failed to create session (probe might not exist)
    - 4 = failed to open probe
    
    @todo Replace use of args from argparse with something cleaner.
    """
    
    def __init__(self, args, cmds=None):
        """! @brief Constructor."""
        # Read command-line arguments.
        self.args = args
        self.cmds = cmds
        
        self.context = CommandExecutionContext(no_init=self.args.no_init)
        self.context.command_set.add_command_group('commander')
        self.session = None
        self.exit_code = 0
        
    def run(self):
        """! @brief Main entry point."""
        try:
            # If no commands, enter interactive mode.
            if self.cmds is None:
                if not self.connect():
                    return self.exit_code
                
                # Print connected message, unless not initing.
                if not self.args.no_init:
                    try:
                        # If the target is locked, we can't read the CPU state.
                        if self.session.target.is_locked():
                            status = "locked"
                        else:
                            try:
                                status = self.session.target.get_state().name.capitalize()
                            except (AttributeError, KeyError):
                                status = "<no core>"

                        # Say what we're connected to.
                        print("Connected to %s [%s]: %s" % (self.context.target.part_number,
                            status, self.session.board.unique_id))
                    except exceptions.TransferFaultError:
                        pass
                else:
                    # Say what we're connected to, but without status.
                    print("Connected to %s [no init mode]: %s" % (self.context.target.part_number,
                        self.session.board.unique_id))

                # Run the REPL interface.
                console = PyocdRepl(self.context)
                console.run()
                
            # Otherwise, run the list of commands we were given and exit. We only connect when
            # there is a command that requires a connection (most do).
            else:
                self.run_commands()

        except ToolExitException:
            self.exit_code = 0
        except exceptions.TransferError:
            print("Error: memory transfer failed")
            # Use get_current() in case our session hasn't been created yet.
            if session.Session.get_current().log_tracebacks:
                traceback.print_exc()
            self.exit_code = 2
        except exceptions.CommandError as e:
            print("Error:", e)
            self.exit_code = 1
        finally:
            # Ensure the session is closed.
            if self.session is not None:
                self.session.close()

        return self.exit_code
    
    def run_commands(self):
        """! @brief Run commands specified on the command line."""
        did_connect = False

        for args in self.cmds:
            # Extract the command name.
            cmd = args[0].lower()
            
            # Handle certain commands without connecting.
            needs_connect = (cmd not in ('list', 'help', 'exit'))

            # For others, connect first.
            if needs_connect and not did_connect:
                if not self.connect():
                    return self.exit_code
                did_connect = True
        
            # Merge commands args back to one string.
            # FIXME this is overly complicated
            cmdline = " ".join('"{}"'.format(a) for a in args)
        
            # Invoke action handler.
            result = self.context.process_command_line(cmdline)
            if result is not None:
                self.exit_code = result
                break

    def connect(self):
        """! @brief Connect to the probe."""
        if (self.args.frequency is not None) and (self.args.frequency != DEFAULT_CLOCK_FREQ_HZ):
            self.context.writei("Setting SWD clock to %d kHz", self.args.frequency // 1000)

        options = convert_session_options(self.args.options)
        
        # Set connect mode. The --connect option takes precedence when set. Then, if --halt is set
        # then the connect mode is halt. If connect_mode is set through -O then use that.
        # Otherwise default to attach.
        if hasattr(self.args, 'connect_mode') and self.args.connect_mode is not None:
            connect_mode = self.args.connect_mode
        elif self.args.halt:
            connect_mode = 'halt'
        elif 'connect_mode' in options:
            connect_mode = None
        else:
            connect_mode = 'attach'
        
        # Connect to board.
        probe = ConnectHelper.choose_probe(
                        blocking=(not self.args.no_wait),
                        unique_id=self.args.unique_id,
                        )
        if probe is None:
            self.exit_code = 3
            return False
        
        # Create a proxy so the probe can be shared between the session and a possible probe server.
        probe_proxy = SharedDebugProbeProxy(probe)
        
        # Create the session.
        self.session = session.Session(probe_proxy,
                        project_dir=self.args.project_dir,
                        config_file=self.args.config,
                        user_script=self.args.script,
                        no_config=self.args.no_config,
                        pack=self.args.pack,
                        target_override=self.args.target_override,
                        connect_mode=connect_mode,
                        frequency=self.args.frequency,
                        options=options,
                        option_defaults=dict(
                            auto_unlock=False,
                            resume_on_disconnect=False,
                            )
                        )
        
        if not self._post_connect():
            self.exit_code = 4
            return False
        
        result = self.context.attach_session(self.session)
        if not result:
            self.exit_code = 1
        return result

    def _post_connect(self):
        """! @brief Finish the connect process.
        
        The session is opened. The `no_init` parameter passed to the constructor determines whether the
        board and target are initialized.
        
        If an ELF file was provided on the command line, it is set on the target.
        
        @param self This object.
        @param session A @ref pyocd.core.session.Session "Session" instance.
        @retval True Session attached and context state inited successfully.
        @retval False An error occurred when opening the session or initing the context state.
        """
        assert self.session is not None
        assert not self.session.is_open
        
        # Open the session.
        try:
            self.session.open(init_board=not self.args.no_init)
        except exceptions.TransferFaultError as e:
            if not self.session.target.is_locked():
                self.context.writei("Transfer fault while initing board: %s", e)
                if self.session.log_tracebacks:
                    self.context.write(traceback.format_exc())
                return False
        except exceptions.Error as e:
            self.context.writei("Exception while initing board: %s", e)
            if self.session.log_tracebacks:
                self.context.write(traceback.format_exc())
            return False

        # Set elf file if provided.
        if self.args.elf:
            self.session.target.elf = os.path.expanduser(self.args.elf)

        # Handle a device with flash security enabled.
        if not self.args.no_init and self.session.target.is_locked():
            self.context.write("Warning: Target is locked, limited operations available. Use 'unlock' "
                                "command to mass erase and unlock, then execute 'reinit'.")
        
        return True




# pyOCD debugger
# Copyright (c) 2006-2013,2018-2019 Arm Limited
# SPDX-License-Identifier: Apache-2.0
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


from enum import Enum


class DAPAccessIntf(object):

    class PORT(Enum):
        """! @brief Physical access ports"""
        DEFAULT = 0
        SWD = 1
        JTAG = 2

    class REG(Enum):
        """! @brief Register for DAP access functions"""
        DP_0x0 = 0
        DP_0x4 = 1
        DP_0x8 = 2
        DP_0xC = 3
        AP_0x0 = 4
        AP_0x4 = 5
        AP_0x8 = 6
        AP_0xC = 7

    class ID(Enum):
        """! @brief Information ID used for call to identify"""
        VENDOR = 1
        PRODUCT = 2
        SER_NUM = 3
        CMSIS_DAP_PROTOCOL_VERSION = 4
        DEVICE_VENDOR = 5
        DEVICE_NAME = 6
        BOARD_VENDOR = 7
        BOARD_NAME = 8
        PRODUCT_FW_VERSION = 9
        CAPABILITIES = 0xf0
        TEST_DOMAIN_TIMER = 0xf1
        UART_RECEIVE_BUFFER_SIZE = 0xfb
        UART_TRANSMIT_BUFFER_SIZE = 0xfc
        SWO_BUFFER_SIZE = 0xfd
        MAX_PACKET_COUNT = 0xfe
        MAX_PACKET_SIZE = 0xff

    class Error(Exception):
        """! @brief Parent of all error DAPAccess can raise"""
        pass

    class DeviceError(Error):
        """! @brief Error communicating with device"""
        pass

    class CommandError(DeviceError):
        """! @brief The host debugger reported failure for the given command"""
        pass

    class TransferError(CommandError):
        """! @brief Error occurred with a transfer over SWD or JTAG"""
        pass

    class TransferTimeoutError(TransferError):
        """! @brief A SWD or JTAG timeout occurred"""
        pass

    class TransferFaultError(TransferError):
        """! @brief A SWD Fault occurred"""
        pass

    class TransferProtocolError(TransferError):
        """! @brief A SWD protocol error occurred"""
        pass

    @staticmethod
    def get_connected_devices():
        """! @brief Return a list of DAPAccess devices"""
        raise NotImplementedError()

    @staticmethod
    def get_device(device_id):
        """! @brief Return the DAPAccess device with the give ID"""
        raise NotImplementedError()

    @staticmethod
    def set_args(arg_list):
        """! @brief Set arguments to configure behavior"""
        raise NotImplementedError()

    @property
    def protocol_version(self):
        """! @brief CMSIS-DAP protocol version.
        
        The version is represented as 3-tuple with elements, in order, of major version,
        minor version, and patch version.

        Not valid (returns None) until the device is opened.
        """
        raise NotImplementedError()

    @property
    def vendor_name(self):
        raise NotImplementedError()

    @property
    def product_name(self):
        raise NotImplementedError()
    
    @property
    def vidpid(self):
        """! @brief A tuple of USB VID and PID, in that order."""
        raise NotImplementedError()

    @property
    def has_swd_sequence(self):
        """! @brief Boolean indicating whether the DAP_SWD_Sequence command is supported.
        
        This property is only valid after the probe is opened. Until then, the value will be None.
        """
        raise NotImplementedError()

    # ------------------------------------------- #
    #          Host control functions
    # ------------------------------------------- #
    def open(self):
        """! @brief Open device and lock it for exclusive access"""
        raise NotImplementedError()

    def close(self):
        """! @brief Close device and unlock it"""
        raise NotImplementedError()

    def get_unique_id(self):
        """! @brief Get the unique ID of this device which can be used in get_device

        This function is safe to call before open is called.
        """
        raise NotImplementedError()

    def identify(self, item):
        """! @brief Return the requested information for this device"""
        raise NotImplementedError()

    # ------------------------------------------- #
    #          Target control functions
    # ------------------------------------------- #
    def connect(self, port=None):
        """! @brief Initialize DAP IO pins for JTAG or SWD"""
        raise NotImplementedError()

    def configure_swd(self, turnaround=1, always_send_data_phase=False):
        """! @brief Modify SWD configuration.
        
        @param self
        @param turnaround Number of turnaround phase clocks, from 1-4.
        @param always_send_data_phase Whether the data phase should always be transmitted on writes,
            even on a FAULT response. This is required for sticky overrun support.
        """
        raise NotImplementedError()
    
    def configure_jtag(self, devices_irlen=None):
        """! @brief Modify JTAG configuration.
        
        @param self
        @param devices_irlen Sequence of IR lengths for each device, thus also specifying the
            number of devices. If not passed, this will default to a single device with IRLen=4.
        """
        raise NotImplementedError()

    def swj_sequence(self, length, bits):
        """! @brief Send sequence to activate JTAG or SWD on the target.
        
        @param self
        @param length Number of bits to transfer on TCK/TMS.
        @param bits Integer with the bit values, sent LSB first.
        """
        raise NotImplementedError()

    def swd_sequence(self, sequences):
        """! @brief Send a sequences of bits on the SWDIO signal.
        
        This method sends the DAP_SWD_Sequence CMSIS-DAP command.
        
        Each sequence in the _sequences_ parameter is a tuple with 1 or 2 members:
        - 0: int: number of TCK cycles from 1-64
        - 1: int: the SWDIO bit values to transfer. The presence of this tuple member indicates the sequence is
            an output sequence; the absence means that the specified number of TCK cycles of SWDIO data will be
            read and returned.
        
        @param self
        @param sequences A sequence of sequence description tuples as described above.
        
        @return A 2-tuple of the response status, and a sequence of bytes objects, one for each input
            sequence. The length of the bytes object is (<TCK-count> + 7) / 8. Bits are in LSB first order.
        """

    def jtag_sequence(self, cycles, tms, read_tdo, tdi):
        """! @brief Send JTAG sequence.
        
        @param self
        @param cycles Number of TCK cycles, from 1-64.
        @param tms Fixed TMS value. Either 0 or 1.
        @param read_tdo Boolean indicating whether TDO should be read.
        @param tdi Integer with the TDI bit values to be transferred each TCK cycle. The LSB is
            sent first.
        
        @return Either an integer with TDI bit values, or None, if _read_tdo_ was false.
        """
        raise NotImplementedError()

    def disconnect(self):
        """! @brief Deinitialize the DAP I/O pins"""
        raise NotImplementedError()

    def set_clock(self, frequency):
        """! @brief Set the frequency for JTAG and SWD in Hz

        This function is safe to call before connect is called.
        """
        raise NotImplementedError()

    def get_swj_mode(self):
        """! @brief Return the current port type - SWD or JTAG"""
        raise NotImplementedError()

    def reset(self):
        """! @brief Reset the target"""
        raise NotImplementedError()

    def assert_reset(self, asserted):
        """! @brief Assert or de-assert target reset line"""
        raise NotImplementedError()
    
    def is_reset_asserted(self):
        """! @brief Returns True if the target reset line is asserted or False if de-asserted"""
        raise NotImplementedError()

    def set_deferred_transfer(self, enable):
        """! @brief Allow reads and writes to be buffered for increased speed"""
        raise NotImplementedError()

    def flush(self):
        """! @brief Write out all unsent commands"""
        raise NotImplementedError()

    def vendor(self, index, data=None):
        """! @brief Send a vendor specific command"""
        raise NotImplementedError()
    
    def has_swo(self):
        """! @brief Returns bool indicating whether the link supports SWO."""
        raise NotImplementedError()

    def swo_configure(self, enabled, rate):
        """! @brief Enable or disable SWO and set the baud rate."""
        raise NotImplementedError()

    def swo_control(self, start):
        """! @brief Pass True to start recording SWO data, False to stop."""
        raise NotImplementedError()

    def get_swo_status(self):
        """! @brief Returns a 2-tuple with a status mask at index 0, and the number of buffered
        SWO data bytes at index 1."""
        raise NotImplementedError()

    def swo_read(self, count=None):
        """! @brief Read buffered SWO data from the target.
        
        The count parameter is optional. If
        provided, it is the number of bytes to read, which must be less than the packet size.
        If count is not provided, the packet size will be used instead.
        
        Returns a 3-tuple containing the status mask at index 0, the number of buffered
        SWO data bytes at index 1, and a list of the received data bytes at index 2."""
        raise NotImplementedError()

    # ------------------------------------------- #
    #          DAP Access functions
    # ------------------------------------------- #
    def write_reg(self, reg_id, value, dap_index=0):
        """! @brief Write a single word to a DP or AP register"""
        raise NotImplementedError()

    def read_reg(self, reg_id, dap_index=0, now=True):
        """! @brief Read a single word to a DP or AP register"""
        raise NotImplementedError()

    def reg_write_repeat(self, num_repeats, reg_id, data_array, dap_index=0):
        """! @brief Write one or more words to the same DP or AP register"""
        raise NotImplementedError()

    def reg_read_repeat(self, num_repeats, reg_id, dap_index=0, now=True):
        """! @brief Read one or more words from the same DP or AP register"""
        raise NotImplementedError()



# pyOCD debugger
# Copyright (c) 2006-2020 Arm Limited
# SPDX-License-Identifier: Apache-2.0
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
from __future__ import print_function

import argparse
import os
import sys
from time import sleep
from random import randrange
import math
import logging

from pyocd.core.helpers import ConnectHelper
from pyocd.core.memory_map import MemoryType
from pyocd.flash.file_programmer import FileProgrammer
from pyocd.utility.conversion import (float32_to_u32, u16le_list_to_byte_list)
from pyocd.utility.mask import same

from test_util import (
    Test,
    get_session_options,
    get_test_binary_path,
    )

# Simple code sequence used to test range stepping.
# The important part is that it has no branches.
RANGE_STEP_CODE = u16le_list_to_byte_list([
        0x3001, # adds    r0, #1
        0x43C1, # mvns    r1, r0
        0x3101, # adds    r1, #1
        0x0102, # movs    r2, r0, lsl #4
        0x0013, # movs    r3, r2
        0x404B, # eors    r3, r1
        0x1840, # adds    r0, r1
        0x1880, # adds    r0, r2
        0x1AC0, # subs    r0, r3
        0xBE00, # bkpt    #0
        ])

class BasicTest(Test):
    def __init__(self):
        super(BasicTest, self).__init__("Basic Test", run_basic_test)
    
def run_basic_test(board_id):
    return basic_test(board_id, None)

def basic_test(board_id, file):
    with ConnectHelper.session_with_chosen_probe(unique_id=board_id, **get_session_options()) as session:
        board = session.board
        addr = 0
        size = 0
        f = None
        binary_file = "l1_"

        if file is None:
            binary_file = get_test_binary_path(board.test_binary)
        else:
            binary_file = file

        print("binary file: %s" % binary_file)

        memory_map = board.target.get_memory_map()
        ram_region = memory_map.get_default_region_of_type(MemoryType.RAM)
        rom_region = memory_map.get_boot_memory()

        addr = ram_region.start
        size = 0x502
        addr_bin = rom_region.start

        target = board.target
        flash = rom_region.flash

        print("\n\n------ GET Unique ID ------")
        print("Unique ID: %s" % board.unique_id)

        print("\n\n------ TEST READ / WRITE CORE REGISTER ------")
        pc = target.read_core_register('pc')
        print("initial pc: 0x%X" % target.read_core_register('pc'))
        # write in pc dummy value
        target.write_core_register('pc', 0x3D82)
        print("now pc: 0x%X" % target.read_core_register('pc'))
        # write initial pc value
        target.write_core_register('pc', pc)
        print("initial pc value rewritten: 0x%X" % target.read_core_register('pc'))

        msp = target.read_core_register('msp')
        psp = target.read_core_register('psp')
        print("MSP = 0x%08x; PSP = 0x%08x" % (msp, psp))

        if 'faultmask' in target.core_registers.by_name:
            control = target.read_core_register('control')
            faultmask = target.read_core_register('faultmask')
            basepri = target.read_core_register('basepri')
            primask = target.read_core_register('primask')
            print("CONTROL = 0x%02x; FAULTMASK = 0x%02x; BASEPRI = 0x%02x; PRIMASK = 0x%02x" % (control, faultmask, basepri, primask))
        else:
            control = target.read_core_register('control')
            primask = target.read_core_register('primask')
            print("CONTROL = 0x%02x; PRIMASK = 0x%02x" % (control, primask))

        target.write_core_register('primask', 1)
        newPrimask = target.read_core_register('primask')
        print("New PRIMASK = 0x%02x" % newPrimask)
        target.write_core_register('primask', primask)
        newPrimask = target.read_core_register('primask')
        print("Restored PRIMASK = 0x%02x" % newPrimask)

        if target.selected_core.has_fpu:
            s0 = target.read_core_register('s0')
            print("S0 = %g (0x%08x)" % (s0, float32_to_u32(s0)))
            target.write_core_register('s0', math.pi)
            newS0 = target.read_core_register('s0')
            print("New S0 = %g (0x%08x)" % (newS0, float32_to_u32(newS0)))
            target.write_core_register('s0', s0)
            newS0 = target.read_core_register('s0')
            print("Restored S0 = %g (0x%08x)" % (newS0, float32_to_u32(newS0)))


        print("\n\n------ TEST HALT / RESUME ------")

        print("resume")
        target.resume()
        sleep(0.2)

        print("halt")
        target.halt()
        print("HALT: pc: 0x%X" % target.read_core_register('pc'))
        sleep(0.2)


        print("\n\n------ TEST STEP ------")

        print("reset and halt")
        target.reset_and_halt()
        currentPC = target.read_core_register('pc')
        print("HALT: pc: 0x%X" % currentPC)
        sleep(0.2)

        for i in range(4):
            print("step")
            target.step()
            newPC = target.read_core_register('pc')
            print("STEP: pc: 0x%X" % newPC)
            sleep(0.2)

        print("\n\n------ TEST RANGE STEP ------")

        # Add some extra room before end of memory, and a second copy so there are instructions
        # after the final bkpt. Add 1 because region end is always odd.
        test_addr = ram_region.end + 1 - len(RANGE_STEP_CODE) * 2 - 32
        # Since the end address is inclusive, we need to exclude the last instruction.
        test_end_addr = test_addr + len(RANGE_STEP_CODE) - 2
        print("range start = %#010x; range_end = %#010x" % (test_addr, test_end_addr))
        # Load up some code into ram to test range step.
        target.write_memory_block8(test_addr, RANGE_STEP_CODE * 2)
        check_data = target.read_memory_block8(test_addr, len(RANGE_STEP_CODE) * 2)
        if not same(check_data, RANGE_STEP_CODE * 2):
            print("Failed to write range step test code to RAM")
        else:
            print("wrote range test step code to RAM successfully")
        
        target.write_core_register('pc', test_addr)
        currentPC = target.read_core_register('pc')
        print("start PC: 0x%X" % currentPC)
        target.step(start=test_addr, end=test_end_addr)
        newPC = target.read_core_register('pc')
        print("end PC: 0x%X" % newPC)

        # Now test again to ensure the bkpt stops it.
        target.write_core_register('pc', test_addr)
        currentPC = target.read_core_register('pc')
        print("start PC: 0x%X" % currentPC)
        target.step(start=test_addr, end=test_end_addr + 4) # include bkpt
        newPC = target.read_core_register('pc')
        print("end PC: 0x%X" % newPC)
        halt_reason = target.get_halt_reason()
        print("halt reason: %s (should be BREAKPOINT)" % halt_reason.name)

        print("\n\n------ TEST READ / WRITE MEMORY ------")
        target.halt()
        print("READ32/WRITE32")
        val = randrange(0, 0xffffffff)
        print("write32 0x%X at 0x%X" % (val, addr))
        target.write_memory(addr, val)
        res = target.read_memory(addr)
        print("read32 at 0x%X: 0x%X" % (addr, res))
        if res != val:
            print("ERROR in READ/WRITE 32")

        print("\nREAD16/WRITE16")
        val = randrange(0, 0xffff)
        print("write16 0x%X at 0x%X" % (val, addr + 2))
        target.write_memory(addr + 2, val, 16)
        res = target.read_memory(addr + 2, 16)
        print("read16 at 0x%X: 0x%X" % (addr + 2, res))
        if res != val:
            print("ERROR in READ/WRITE 16")

        print("\nREAD8/WRITE8")
        val = randrange(0, 0xff)
        print("write8 0x%X at 0x%X" % (val, addr + 1))
        target.write_memory(addr + 1, val, 8)
        res = target.read_memory(addr + 1, 8)
        print("read8 at 0x%X: 0x%X" % (addr + 1, res))
        if res != val:
            print("ERROR in READ/WRITE 8")


        print("\n\n------ TEST READ / WRITE MEMORY BLOCK ------")
        data = [randrange(1, 50) for x in range(size)]
        target.write_memory_block8(addr, data)
        block = target.read_memory_block8(addr, size)
        error = False
        for i in range(len(block)):
            if (block[i] != data[i]):
                error = True
                print("ERROR: 0x%X, 0x%X, 0x%X!!!" % ((addr + i), block[i], data[i]))
        if error:
            print("TEST FAILED")
        else:
            print("TEST PASSED")


        print("\n\n------ TEST RESET ------")
        target.reset()
        sleep(0.1)
        target.halt()

        for i in range(5):
            target.step()
            print("pc: 0x%X" % target.read_core_register('pc'))

        print("\n\n------ TEST PROGRAM/ERASE PAGE ------")
        # Fill 3 pages with 0x55
        sector_size = rom_region.sector_size
        page_size = rom_region.page_size
        sectors_to_test = min(rom_region.length // sector_size, 3)
        addr_flash = rom_region.start + rom_region.length - sector_size * sectors_to_test
        fill = [0x55] * page_size
        for i in range(0, sectors_to_test):
            address = addr_flash + sector_size * i
            # Test only supports a location with 3 aligned
            # pages of the same size
            current_page_size = flash.get_page_info(addr_flash).size
            assert page_size == current_page_size
            assert address % current_page_size == 0

            print("Erasing sector @ 0x%x (%d bytes)" % (address, sector_size))
            flash.init(flash.Operation.ERASE)
            flash.erase_sector(address)

            print("Verifying erased sector @ 0x%x (%d bytes)" % (address, sector_size))
            data = target.read_memory_block8(address, sector_size)
            if data != [flash.region.erased_byte_value] * sector_size:
                print("FAILED to erase sector @ 0x%x (%d bytes)" % (address, sector_size))
            else:
                print("Programming page @ 0x%x (%d bytes)" % (address, page_size))
                flash.init(flash.Operation.PROGRAM)
                flash.program_page(address, fill)

                print("Verifying programmed page @ 0x%x (%d bytes)" % (address, page_size))
                data = target.read_memory_block8(address, page_size)
                if data != fill:
                    print("FAILED to program page @ 0x%x (%d bytes)" % (address, page_size))

        # Erase the middle sector
        if sectors_to_test > 1:
            address = addr_flash + sector_size
            print("Erasing sector @ 0x%x (%d bytes)" % (address, sector_size))
            flash.init(flash.Operation.ERASE)
            flash.erase_sector(address)
        flash.cleanup()

        print("Verifying erased sector @ 0x%x (%d bytes)" % (address, sector_size))
        data = target.read_memory_block8(address, sector_size)
        if data != [flash.region.erased_byte_value] * sector_size:
            print("FAILED to erase sector @ 0x%x (%d bytes)" % (address, sector_size))
       
        # Re-verify the 1st and 3rd page were not erased, and that the 2nd page is fully erased
        did_pass = False
        for i in range(0, sectors_to_test):
            address = addr_flash + sector_size * i
            print("Verifying page @ 0x%x (%d bytes)" % (address, page_size))
            data = target.read_memory_block8(address, page_size)
            expected = ([flash.region.erased_byte_value] * page_size) if (i == 1) else fill
            did_pass = (data == expected)
            if not did_pass:
                print("FAILED verify for page @ 0x%x (%d bytes)" % (address, page_size))
                break
        if did_pass:
            print("TEST PASSED")
        else:
            print("TEST FAILED")

        print("\n\n----- FLASH NEW BINARY -----")
        FileProgrammer(session).program(binary_file, base_address=addr_bin)

        target.reset()

if __name__ == "__main__":
    parser = argparse.ArgumentParser(description='A CMSIS-DAP python debugger')
    parser.add_argument('-f', help='binary file', dest="file")
    parser.add_argument('-d', '--debug', action="store_true", help='Enable debug logging')
    args = parser.parse_args()
    level = logging.DEBUG if args.debug else logging.INFO
    logging.basicConfig(level=level)
    file = args.file
    basic_test(None, file)



# pyOCD debugger
# Copyright (c) 2021 Chris Reed
# SPDX-License-Identifier: Apache-2.0
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import re
from typing import (Iterable, Sequence, Optional, Tuple)

class UniquePrefixMatcher:
    """! @brief Manages detection of shortest unique prefix match of a set of strings."""
    
    def __init__(self, items: Optional[Iterable[str]] = None):
        """! @brief Constructor.
        @param self This object.
        @param items Optional sequence of strings.
        """
        self._items = set(items) if (items is not None) else set()
    
    def add_items(self, items: Iterable[str]) -> None:
        """! @brief Add some items to be matched.
        @param self This object.
        @param items Sequence of strings.
        """
        self._items.update(items)
    
    def find_all(self, prefix: str) -> Tuple[str, ...]:
        """! @brief Return all items matching the given prefix.
        @param self This object.
        @param prefix String that is compared as a prefix against the items passed to the constructor.
            Must not be the empty string.
        @return List of all items where `prefix` is a valid prefix.
        @exception ValueError Raised for an empty `prefix`.
        """
        if len(prefix) == 0:
            raise ValueError("empty prefix")
        # First look for an exact match.
        if prefix in self._items:
            return (prefix,)
        return tuple(i for i in self._items if i.startswith(prefix))
    
    def find_one(self, prefix: str) -> Optional[str]:
        """! @brief Return the item matching the given prefix, or None.
        @param self This object.
        @param prefix String that is compared as a prefix against the items passed to the constructor.
        @return The full value of the matching item where `prefix` is a valid prefix.
        @exception ValueError Raised for an empty `prefix`.
        """
        all_matches = self.find_all(prefix)
        if len(all_matches) == 1:
            return all_matches[0]
        return None


_INT_SUFFIX_RE = re.compile(r'[0-9]+$')

def uniquify_name(name: str, others: Sequence[str]) -> str:
    """@brief Ensure the given name is unique amongst the other provided names.
    
    If the `name` parameter is not unique, an integer will be appended to it. If the name already ends in an
    integer, that value will be incremented by 1.
    
    @param name The name to uniqify.
    @param others Sequence of other names to compare against.
    @return A string guaranteed to not be the same as any string contained in `others`.
    """
    while name in others:
        # Look for an integer at the end.
        matches = list(_INT_SUFFIX_RE.finditer(name))
        if len(matches):
            match = matches[0]
            u_value = int(match.group())
            name = name[:match.start()]
        else:
            name += "_"
            u_value = 0

        # Update the name with the trailing int incremented.
        name += str(u_value + 1)
    
    return name



# pyOCD debugger
# Copyright (c) 2020 Arm Limited
# SPDX-License-Identifier: Apache-2.0
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
from __future__ import print_function

import argparse
import sys
import traceback
import logging
import six
import os
from collections import UserDict

from pyocd.core.helpers import ConnectHelper
from pyocd.probe.pydapaccess import DAPAccess
from pyocd.utility.mask import round_up_div
from pyocd.utility import conversion
from pyocd.core.memory_map import MemoryType
from pyocd.commands.commander import PyOCDCommander
from pyocd.commands import commands
from test_util import (
    Test,
    TestResult,
    get_session_options,
    binary_to_elf_file,
    PYOCD_DIR,
    )

GDB_TEST_ELF = os.path.join(PYOCD_DIR, "src/gdb_test_program/gdb_test.elf")

class CommanderTestResult(TestResult):
    def __init__(self):
        super(CommanderTestResult, self).__init__(None, None, None)
        self.name = "commander"

class CommanderTest(Test):
    def __init__(self):
        super(CommanderTest, self).__init__("Commander Test", commander_test)

    def run(self, board):
        try:
            result = self.test_function(board.unique_id)
        except Exception as e:
            result = CommanderTestResult()
            result.passed = False
            print("Exception %s when testing board %s" % (e, board.unique_id))
            traceback.print_exc(file=sys.stdout)
        result.board = board
        result.test = self
        return result

def commander_test(board_id):
    test_pass_count = 0
    test_count = 0
    failed_commands = []
    result = CommanderTestResult()
    
    COMMANDS_TO_TEST = [
            # general commands
            ["continue"],
            ["status"],
            ["halt"],
            ["status"],
            
            # commander command group - these are not tested by commands_test.py.
            ["list"],
            ["exit"], # Must be last command!
            ]

    print("\n------ Testing commander ------\n")
    
    # Set up commander args.
    args = UserDict()
    args.no_init = False
    args.frequency = 1000000
    args.options = {} #get_session_options()
    args.halt = True
    args.no_wait = True
    args.project_dir = None
    args.config = None
    args.script = None
    args.no_config = False
    args.pack = None
    args.unique_id = board_id
    args.target_override = None
    args.elf = GDB_TEST_ELF
    
    test_count += 1
    try:
        cmdr = PyOCDCommander(args, COMMANDS_TO_TEST)
        cmdr.run()
        test_pass_count += 1
        print("TEST PASSED")
    except Exception:
        print("TEST FAILED")
        traceback.print_exc()
        
    test_count += 1
    print("Testing exit code")
    print("Exit code:", cmdr.exit_code)
    if cmdr.exit_code == 0:
        test_pass_count += 1
        print("TEST PASSED")
    else:
        print("TEST FAILED")

    print("\n\nTest Summary:")
    print("Pass count %i of %i tests" % (test_pass_count, test_count))
    if failed_commands:
        for c in failed_commands:
            print(" - '" + c + "'")
    if test_pass_count == test_count:
        print("COMMANDER TEST SCRIPT PASSED")
    else:
        print("COMMANDER TEST SCRIPT FAILED")

    result.passed = test_count == test_pass_count
    return result

if __name__ == "__main__":
    parser = argparse.ArgumentParser(description='pyOCD commander test')
    parser.add_argument('-d', '--debug', action="store_true", help='Enable debug logging')
    parser.add_argument('-u', '--uid', help='Debug probe unique ID')
    parser.add_argument("-da", "--daparg", dest="daparg", nargs='+', help="Send setting to DAPAccess layer.")
    args = parser.parse_args()
    level = logging.DEBUG if args.debug else logging.INFO
    logging.basicConfig(level=level)
    DAPAccess.set_args(args.daparg)
    commander_test(args.uid)




# pyOCD debugger
# Copyright (c) 2006-2019 Arm Limited
# Copyright (c) 2021 Chris Reed
# SPDX-License-Identifier: Apache-2.0
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from enum import Enum

from .memory_interface import MemoryInterface
from .memory_map import MemoryMap

class Target(MemoryInterface):

    class State(Enum):
        """! @brief States a target processor can be in."""
        ## Core is executing code.
        RUNNING = 1
        ## Core is halted in debug mode.
        HALTED = 2
        ## Core is being held in reset.
        RESET = 3
        ## Core is sleeping due to a wfi or wfe instruction.
        SLEEPING = 4
        ## Core is locked up.
        LOCKUP = 5
    
    class SecurityState(Enum):
        """! @brief Security states for a processor with the Security extension."""
        ## PE is in the Non-secure state.
        NONSECURE = 0
        ## PE is in the Secure state.
        SECURE = 1

    class ResetType(Enum):
        """! @brief Available reset methods."""
        ## Hardware reset via the nRESET signal.
        HW = 1
        ## Software reset using the core's default software reset method.
        SW = 2
        ## Software reset using the AIRCR.SYSRESETREQ bit.
        SW_SYSRESETREQ = 3
        ## Software reset the entire system (alias of #SW_SYSRESETREQ).
        SW_SYSTEM = SW_SYSRESETREQ
        ## Software reset using the AIRCR.VECTRESET bit.
        #
        # v6-M and v8-M targets do not support VECTRESET, so they will fall back to SW_EMULATED.
        SW_VECTRESET = 4
        ## Software reset the core only (alias of #SW_VECTRESET).
        SW_CORE = SW_VECTRESET
        ## Emulated software reset.
        SW_EMULATED = 5

    class BreakpointType(Enum):
        """! @brief Types of breakpoints."""
        ## Hardware breakpoint.
        HW = 1
        ## Software breakpoint.
        SW = 2
        ## Auto will select the best type given the address and available breakpoints.
        AUTO = 3

    class WatchpointType(Enum):
        """! @brief Types of watchpoints."""
        ## Watchpoint on read accesses.
        READ = 1
        ## Watchpoint on write accesses.
        WRITE = 2
        ## Watchpoint on either read or write accesses.
        READ_WRITE = 3

    class VectorCatch:
        """! Vector catch option masks.
        
        These constants can be OR'd together to form any combination of vector catch settings.
        """
        ## Disable vector catch.
        NONE = 0
        ## Trap on HardFault exception.
        HARD_FAULT = (1 << 0)
        ## Trap on BusFault exception.
        BUS_FAULT = (1 << 1)
        ## Trap on MemManage exception.
        MEM_FAULT = (1 << 2)
        ## Trap on fault occurring during exception entry or exit.
        INTERRUPT_ERR = (1 << 3)
        ## Trap on UsageFault exception caused by state information error, such as an undefined
        # instruction exception.
        STATE_ERR = (1 << 4)
        ## Trap on UsageFault exception caused by checking error, for example an alignment check error.
        CHECK_ERR = (1 << 5)
        ## Trap on UsageFault exception caused by a failed access to a coprocessor.
        COPROCESSOR_ERR = (1 << 6)
        ## Trap on local reset.
        CORE_RESET = (1 << 7)
        ## Trap SecureFault.
        SECURE_FAULT = (1 << 8)
        ALL = (HARD_FAULT | BUS_FAULT | MEM_FAULT | INTERRUPT_ERR
                    | STATE_ERR | CHECK_ERR | COPROCESSOR_ERR | CORE_RESET
                    | SECURE_FAULT)

    class Event(Enum):
        """! Target notification events."""
        ## Sent after completing the initialisation sequence.
        POST_CONNECT = 1
        ## Sent prior to disconnecting cores and powering down the DP.
        PRE_DISCONNECT = 2
        ## Sent prior to resume or step.
        #
        # Associated data is a RunType enum.
        PRE_RUN = 3
        ## Sent after a resume or step operation.
        #
        # For resume, this event will be sent while the target is still running. Use a halt event
        # to trap when the target stops running.
        #
        # Associated data is a RunType enum.
        POST_RUN = 4
        ## Sent prior to a user-invoked halt.
        #
        # Associated data is a HaltReason enum, which will currently always be HaltReason.USER.
        PRE_HALT = 5
        ## Sent after the target halts.
        #
        # Associated data is a HaltReason enum.
        POST_HALT = 6
        ## Sent before executing a reset operation.
        PRE_RESET = 7
        ## Sent after the target has been reset.
        POST_RESET = 8
        ## Sent before programming target flash.
        PRE_FLASH_PROGRAM = 9
        ## Sent after target flash has been reprogrammed.
        POST_FLASH_PROGRAM = 10

    class RunType(Enum):
        """! Run type for run notifications.
        
        An enum of this type is set as the data attribute on PRE_RUN and POST_RUN notifications.
        """
        ## Target is being resumed.
        RESUME = 1
        ## Target is being stepped one instruction.
        STEP = 2

    class HaltReason(Enum):
        """! Halt type for halt notifications.
        
        An value of this type is returned from Target.get_halt_reason(). It is also used as the data
        attribute on PRE_HALT and POST_HALT notifications.
        """
        ## Target halted due to user action.
        USER = 1
        ## Target halted because of a halt or step event.
        DEBUG = 2
        ## Breakpoint event.
        BREAKPOINT = 3
        ## DWT watchpoint event.
        WATCHPOINT = 4
        ## Vector catch event.
        VECTOR_CATCH = 5
        ## External debug request.
        EXTERNAL = 6
        ## PMU event. v8.1-M only.
        PMU = 7

    def __init__(self, session, memory_map=None):
        self._session = session
        self._delegate = None
        # Make a target-specific copy of the memory map. This is safe to do without locking
        # because the memory map may not be mutated until target initialization.
        self.memory_map = memory_map.clone() if memory_map else MemoryMap()
        self._svd_location = None
        self._svd_device = None

    @property
    def session(self):
        return self._session
    
    @property
    def delegate(self):
        return self._delegate
    
    @delegate.setter
    def delegate(self, the_delegate):
        self._delegate = the_delegate
    
    def delegate_implements(self, method_name):
        return (self._delegate is not None) and (hasattr(self._delegate, method_name))
    
    def call_delegate(self, method_name, *args, **kwargs):
        if self.delegate_implements(method_name):
            return getattr(self._delegate, method_name)(*args, **kwargs)
        else:
            # The default action is always taken if None is returned.
            return None

    @property
    def svd_device(self):
        return self._svd_device
    
    @property
    def supported_security_states(self):
        raise NotImplementedError()
    
    @property
    def core_registers(self):
        raise NotImplementedError()

    def is_locked(self):
        return False

    def create_init_sequence(self):
        raise NotImplementedError()

    def init(self):
        raise NotImplementedError()

    def disconnect(self, resume=True):
        pass

    def flush(self):
        self.session.probe.flush()

    def halt(self):
        raise NotImplementedError()

    def step(self, disable_interrupts=True, start=0, end=0, hook_cb=None):
        raise NotImplementedError()

    def resume(self):
        raise NotImplementedError()

    def mass_erase(self):
        raise NotImplementedError()

    def read_core_register(self, id):
        raise NotImplementedError()

    def write_core_register(self, id, data):
        raise NotImplementedError()

    def read_core_register_raw(self, reg):
        raise NotImplementedError()

    def read_core_registers_raw(self, reg_list):
        raise NotImplementedError()

    def write_core_register_raw(self, reg, data):
        raise NotImplementedError()

    def write_core_registers_raw(self, reg_list, data_list):
        raise NotImplementedError()

    def find_breakpoint(self, addr):
        raise NotImplementedError()

    def set_breakpoint(self, addr, type=BreakpointType.AUTO):
        raise NotImplementedError()

    def get_breakpoint_type(self, addr):
        raise NotImplementedError()

    def remove_breakpoint(self, addr):
        raise NotImplementedError()

    def set_watchpoint(self, addr, size, type):
        raise NotImplementedError()

    def remove_watchpoint(self, addr, size, type):
        raise NotImplementedError()

    def reset(self, reset_type=None):
        raise NotImplementedError()

    def reset_and_halt(self, reset_type=None):
        raise NotImplementedError()

    def get_state(self):
        raise NotImplementedError()
        
    def get_security_state(self):
        raise NotImplementedError()

    def get_halt_reason(self):
        raise NotImplementedError()

    @property
    def run_token(self):
        return 0

    def is_running(self):
        return self.get_state() == Target.State.RUNNING

    def is_halted(self):
        return self.get_state() == Target.State.HALTED

    def get_memory_map(self):
        return self.memory_map

    def set_vector_catch(self, enableMask):
        raise NotImplementedError()

    def get_vector_catch(self):
        raise NotImplementedError()

    def get_target_context(self, core=None):
        raise NotImplementedError()



# pyOCD debugger
# Copyright (c) 2015-2020 Arm Limited
# Copyright (c) 2021 Chris Reed
# SPDX-License-Identifier: Apache-2.0
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import threading
import logging
import pkg_resources
import zipfile

from .parser import SVDParser

LOG = logging.getLogger(__name__)

## Path within the pyocd package to the generated zip containing builting SVD files.
BUILTIN_SVD_DATA_PATH = "debug/svd/svd_data.zip"

class SVDFile(object):
    @classmethod
    def from_builtin(cls, svd_name):
        try:
            zip_stream = pkg_resources.resource_stream("pyocd", BUILTIN_SVD_DATA_PATH)
            zip = zipfile.ZipFile(zip_stream, 'r')
            return SVDFile(zip.open(svd_name))
        except (KeyError, FileNotFoundError, zipfile.BadZipFile) as err:
            from ...core.session import Session
            LOG.warning("unable to open builtin SVD file: %s", err, exc_info=Session.get_current().log_tracebacks)
            return None
    
    def __init__(self, filename=None):
        self.filename = filename
        self.device = None

    def load(self):
        self.device = SVDParser.for_xml_file(self.filename).get_device()

class SVDLoader(threading.Thread):
    """! @brief Thread to read an SVD file in the background."""

    def __init__(self, svdFile, completionCallback):
        super(SVDLoader, self).__init__(name='load-svd')
        self.daemon = True
        self._svd_location = svdFile
        self._svd_device = None
        self._callback = completionCallback

    @property
    def device(self):
        if not self._svd_device:
            self.join()
        return self._svd_device

    def load(self):
        if not self._svd_device and self._svd_location:
            self.start()

    def run(self):
        try:
            self._svd_location.load()
            self._svd_device = self._svd_location.device
            if self._callback:
                self._callback(self._svd_device)
        except IOError:
            LOG.warning("Failed to load SVD file %s", self._svd_location.filename)
